;-----------------------------------------------------------------------------
; Z9001-BIOS fr CPA, 1989 Dr. Schwarzenberg
; reass 2004/2006/2008 Volker Pohlers
; Version cpmz6010 (ROSSENDORF)
;
; Das Bios ist recht umfangreich und umfasst Gertetreiber fr ...
; und fr 2 Floppylaufwerke 800K, die am Robotron-Floppy-Modul angeschlossen
; sind. Desweiteren wird unbedingt ein spezielles 64K-RAM-Modul bentigt, da
; das Bios den 2. Speicherbereich 4000H-7FFFH fr sich benutzt.
;
; Speicherbelegung (CCP auf cc00)
;
; 4000-7FFF 	(2. Bank) Shadow Bios 4000-55FF CCP+BDOS-Kopie, 5600..Start-1 SBIOS, danach IO-Buffer fr FDC
; cc00-d3ff	CPA-CCP
; d400-e1ff	BDOS
; e200-e7ff	Upper Bios + bios Ram
;
; Ablauf des CPM-Starts
;
; 1. vom Boot-Loader wird @cpmz9 (init+ccp+bdos+sbios+binit+ubios) nach 8000h geladen
;    und auf 8000H+1600H = 9600H gestartet. Das ist die initiale Adr. von SBIOS,
;    deshalb steht hier ein JP 9FF0h
; 2. init transportiert ccp+bdos+sbios+binit+ubios nach 4000h in die 2. Speicherbank
; 3. binit wird nach 180h geladen, ubios nach e200h geladen.
; 3. binit wird gestartet und initialisiert die Gertetreiber, testet den Speicher,
;    gibt die Startmeldung aus, kopiert einen Startbefehl in den CCP-Befehlspuffer
;    und springt zu WBOOT im Shadow Bios.
; 4. wboot kopiert CCP, (damit wird in cpmz9ok obiger Startbefehl wieder berschrieben...)
;    initialisiert CPA-RAM 0000h ff. und startet schlussendlich den CCP.
; 5. der Speicher ab Start=fdcbuffer in Shadow-Ram wird als IO-Buffer der FDC-Operationen genutzt,
;    die Teile binit und ubios werden also im Shadow-Ram berschrieben
;
; Verbindung zum OS
;
; das ganze Bios ist recht stark mit dem Z9001-OS verknpft (Gertetreiber und
; Gertetreiberverwaltung).
; Es luft daher warscheinlich nicht auf modifizierten Bios'en.
;
; s. auch 00cpm.txt, KCCPM.TXT, README.TXT
;
; Linkeraufruf?  N,/X,/P:D980,CPMLDR,CCP,BDOS,BIOZSW/E /E BIOS
;-----------------------------------------------------------------------------


		cpu	z80
		page	0

		title	"CPMZ9 BIOS"


BIOSVER 	EQU	'CPMZ6010'	; Version CPMZ6010


	IF BIOSVER = 'CPMZ9OK'
FDCPORTS	EQU 	'ROBOTRON'	; Ports 98h, 99h, 0A0h
	ELSEIF BIOSVER = 'CPMSS4R'
FDCPORTS	EQU 	'ROBOTRON'	; Ports 98h, 99h, 0A0h
	ELSE
FDCPORTS	EQU 	'ROSSENDORF'	; Ports 10h, 11h, 12h
	ENDIF


; vp 24.10.2009 
; Das BIOS entspricht CPMSD8, enthlt aber RDR-, PUN- und LST-Treiber fr die S6010.
; Die S6010 wird vermutlich ber ein normales V24-Modul mit anderen Adressen
; angeschlossen und dient als Tastatur und Drucker

SIODA		equ	20h		; SIODA SIO-Kanal Daten
SIOKD		equ	22h		; SIOKD SIO-Kanal Kommandos
CTC		equ	24h		; Adr CTC Kanal 0

;-----------------------------------------------------------------------------
;
;-----------------------------------------------------------------------------

;
CCP		equ	0CC00h		; base of ccp
BDOS       	equ	CCP+0806h	; base of bdos
UBIOS		equ	CCP+1600h	; base of bios

SHADOWRAM	equ	4000H		; Adresse Shadow-Memory fr CCP-Kopie und shadow bios
LOADADRESS	equ	8000H		; Hierhin wird das Programm durch den Loader geladen
SBIOS		equ	SHADOWRAM+1600h	; 5600H

;Ports
		IF BIOSVER = 'CPMZ9OK'

port71h		equ	71h	; RAM-Modul U. Zander, toggle w/o <-> r/w
				; ebenso Lesen von 0F800h (BOOT-Modul zuschalten)
rambank1        equ	72h  	; ShadowRAM deaktivieren; 	vp 04h
rambank2        equ	73h  	; ShadowRAM aktivieren;		vp 05h

		ELSE

rambank1        equ	04h  	; ShadowRAM deaktivieren
rambank2        equ	05h  	; ShadowRAM aktivieren
port71h		equ	06h	; RAM-Modul write/only

		ENDIF

CTC0		equ	80h	; System CTC0
CTC1		equ	81h	; System CTC1
CTC3		equ	83h	; System CTC3

PIO1AD		equ	88h	; System PIO1AD
PIO1BD		equ	89h	; System PIO1BD	Anwenderport
PIO1BC		equ	8bh	; System PIO1BC Anwenderport


; Funktionen fr AS
hi	function x,(x>>8)&255 	; oberes Byte eines 16-Bit-Wortes


;-----------------------------------------------------------------------------
; Initialisierung Loader (CPMLDR)
; vom aktuellen Boot-Loader wird dieser Code nicht ausgefhrt, es wird nur
; die Ladeadresse entnommen.
;-----------------------------------------------------------------------------
;
		org 	SBIOS - 1600H - 128

		di
		ld	de, LOADADRESS	; <- dieser Befehl mu stehen, damit der
					; Loader das CPMZ9 als gltig erkennt
					; Gleichzeitig entnimmt der Loader hier
					; die initiale Ladeadresse.
		ld	hl, 180h
                ld      bc, romend-SHADOWRAM
		ld	de, SHADOWRAM	; ab hier 1600h Byte CCP+BDOS, dann BIOS (auf 5600H)
	IF BIOSVER = 'CPMZ6010'
		in	a, (68h)	; ????
	ELSE
		in	a, (rambank2)	; ShadowRAM aktivieren
	ENDIF
		ldir
		xor	a
		jp	start1

;-----------------------------------------------------------------------------
; CCP+BDOS
;-----------------------------------------------------------------------------

		org	SBIOS - 1600H


		IF BIOSVER = 'CPMZ9OK'
		binclude ccp1cc00.cpm
		ELSEIF CCP = 0CC00h
		binclude ccp2cc00.cpm
		ELSEIF CCP = 0CD00h
		binclude ccp2cd00.cpm
		ELSE
		db	1600h dup (?)	; Bereich CCP+BDOS
		ENDIF

;-----------------------------------------------------------------------------
; Shadow BIOS
;-----------------------------------------------------------------------------

		org 	SBIOS

		jp	start+LOADADRESS-SHADOWRAM	; boot im Loadbereich 8000h ff.
		jp	wboot
		jp	_const		; const im upper bios
		jp	0F009h		; conin im z9001-os
		jp	_conout		; conout im upper bios
		jp	0F00Fh		; list im z9001-os
		jp	0F012h		; punch im z9001-os
		jp	0F015h		; reader im z9001-os
		jp	home
		jp	seldsk
		jp	settrk
		jp	setsec
		jp	setdma
uread:		jp	read
uwrite:		jp	write
		jp	0F02Dh		; listst im z9001-os
		jp	sectran


		IF BIOSVER <> 'CPMZ9OK'
		db	"CPM9Z"
		ENDIF

;-----------------------------------------------------------------------------
; PUN-Gertetreiber fr	PUNCH
;-----------------------------------------------------------------------------
upunch:		ld	a, c
		ld	de, upunch1+1
		call	hexa		; A nach hex konvertieren, Eintragen nach (DE),	2x inc DE
		ld	c, '('
		call	_conout		; _conout im upper Bios
upunch1:	ld	bc, 0
		push	bc
		call	_conout		; _conout im upper Bios
		pop	bc
		ld	c, b
		call	_conout		; _conout im upper Bios
		ld	c, ')'
		jp	_conout		; _conout im upper Bios
stopkey:	call	0FD33h		; Z9001-OS DECO0: Abfrage Tastaturmatrix
		ei
		sub	3		; STOP-Taste ?
		or	a
		ret	nz
		ld	(25h), a	; Tastaturbuffer
		scf
		ret

;-----------------------------------------------------------------------------
; TTY-Treiber fr CONS
;-----------------------------------------------------------------------------
uttyc:		cp	1		;
uttyc1:		scf
		ccf
		jp	nz, 0F8F1h	; Z9001-OS CRT:	Steuerprogramm des CRT-Treibers
uttyc2:		call	0F8F1h		; Z9001-OS CRT:	Steuerprogramm des CRT-Treibers
		call	bell
		ld	hl, lstflag

		IF BIOSVER <> 'CPMSS4R'

		cp	1Ch		; LIST-Taste
		jr	nz, uttyc3
		ld	a, 1
		ld	(hl), a
		jr	uttyc2
uttyc3:		bit	0, (hl)
		jr	z, uttyc4
		ld	(hl), 0

		ELSEIF BIOSVER = 'CPMSS4R'

		bit	0, (hl)
		jr	nz, uttyc10
		cp	1Ch		; LIST-Taste
		jr	nz, uttyc10
		ld	a, 1
		ld	(hl), a
		jr	uttyc

uttyc10:	bit	0, (hl)
		jr	z, uttyc4
		ld	(hl), 0
		cp	1Ch		; LIST-Taste
		ret	z
		cp	'P'
		jr	nz, uttyc12
		ld	a, (15h)
		xor	1
		ld	(15h), a
uttyc11:	ld	a, 1
		jr	uttyc

uttyc12:	cp	'N'
		jr	nz, uttyc18
		push	de
		ld	hl, 0EC00h	; Adr. Bildwiederholspeicher
		in	a, (PIO1AD)
		bit	2, a
		ld	a, 14h
		jr	nz, uttyc13
		add	a, 4
uttyc13:	ld	d, a
uttyc14:	ld	e, 28h ; '('
uttyc15:	ld	c, (hl)
		push	de
		push	hl
		call	0F00Fh		; Z9001-OS: list
		pop	hl
		pop	de
		jr	c, uttyc17
		inc	hl
		dec	e
		jr	nz, uttyc15
		push	hl
		push	de
		ld	c, 0Dh
		call	0F00Fh		; Z9001-OS: list
		jr	c, uttyc16
		ld	c, 0Ah
		call	0F00Fh		; Z9001-OS: list
uttyc16:	pop	de
		pop	hl
		jr	c, uttyc17
		dec	d
		jr	nz, uttyc14
uttyc17:	pop	de
		jr	uttyc11
uttyc18:	equ	$
		ENDIF

		ld	hl, lsttab	; Tabelle Extrazeichen (LIST+char)
		ld	bc, 7
		cpir			; suche	Sondertaste
		jr	nz, uttyc4
		ld	bc, 6
		add	hl, bc
		ld	a, (hl)		; Sonderzeichen	holen

uttyc4:		ld	hl, uttyc6
		push	hl
		cp	'A'
		jr	c, uttyc5
		cp	'['
		jr	nc, uttyc5
		or	20h
		ret
uttyc5:		cp	'a'
		ret	c
		cp	'{'
		ret	nc
		sub	20h
		ret
uttyc6:	or	a
		ret
lstflag:	db	0
lsttab:		db	'8'		; Tabelle Extrazeichen (LIST+char)
		db	'9'
		db	','
		db	'.'
		db	'I'
		db	'?'
		IF BIOSVER = 'CPMSS4R'
		db	'='
		ELSE
		db	's'
		ENDIF
;
		db	'['
		db	']'
		db	'{'
		db	'}'
		db	'|'
		db	5Ch		; '\'
		db	'~'

;-----------------------------------------------------------------------------
; CRT-Treiber fr CONS
;-----------------------------------------------------------------------------
ucrtc:		cp	1		;
		IF BIOSVER <> 'CPMSS4R'
		jr	nz, uttyc1
		ELSEIF BIOSVER = 'CPMSS4R'
		jp	nz, uttyc1
		ENDIF
		call	0F8F1h		; Z9001-OS CRT:	Steuerprogramm des CRT-Treibers

bell:		di
		push	af
		push	bc
		ld	b, 0
		ld	c, 14h
		ld	a, 00000111b	; interrupt aus, zeitgeber mode, Vorteiler 16, negative	Flange,
					; Start	sofort,	Konstante folgt, Kanal Reset
		out	(CTC0), a	; CTC0
		ld	a, 96h		; Zeitkonstante
		out	(CTC0), a
		in	a, (PIO1AD)
		set	7, a
		out	(PIO1AD), a	; Lautsprecher an
bell1:		djnz	bell1
		dec	c
		jr	nz, bell1
		res	7, a
		out	(PIO1AD), a	; Lautsprecher aus
		ld	a, 00000011b
		out	(CTC0), a	; CTC0 Reset
		pop	bc
		pop	af
		ei
		ret

;-----------------------------------------------------------------------------
; SBIOS S6010-Treiber f. PUNCH
;-----------------------------------------------------------------------------

UP6010:		or	a
		jr	nz, UP6010_1
		in	a, (SIOKD)
		and	4
		ret
;
UP6010_1:	dec	a
		dec	a
		scf
		ret	nz
;
UP6010_2:	in	a, (SIOKD)
		bit	2, a		; Sender Puffer leer ?
		jr	nz, UP6010_3
		call	stopkey
		ret	c
		jr	UP6010_2
;
UP6010_3:	ld	a, (UR6010_b1)
		cp	1
		jr	nz, UP6010_4
		set	7, c
UP6010_4:	ld	a, c
		cp	2
		jr	nz, UP6010_5
		ld	a, 8
UP6010_5:	cp	1
		ld	(UR6010_b1), a
		ret	z
		cp	7Fh
		jr	nz, UP6010_6
		ld	a, 1Bh
UP6010_6:	out	(SIODA), a
		ld	hl, 1F00h
;
UP6010_7:	dec	hl
		ld	a, h
		or	l
		jr	nz, UP6010_7
		or	a
		ret

;-----------------------------------------------------------------------------
; SBIOS S6010-Treiber f. READER
;-----------------------------------------------------------------------------

UR6010:		or	a
		jr	nz, UR6010_1
		call	UR6010_u1
		ld	hl, 12Ch	; ???
		call	UP6010_7
		call	UR6010_u2
		in	a, (SIOKD)
		and	1		; Zeichen im Empfngerpuffer ?
		ret

UR6010_1:	dec	a
		scf
		ret	nz

		ld	hl, UR6010_b2
		xor	a
		cp	(hl)
		jr	z, UR6010_2
		ld	a, (hl)
		ld	(hl), 0
		or	a
		ret

UR6010_2:	call	UR6010_u1
UR6010_3:	in	a, (SIOKD)
		rrca
		jr	c, UR6010_4
		call	stopkey
		ld	a, 1Ah
		jr	c, UR6010_8
		jr	UR6010_3
UR6010_4:	ld	a, 1
		out	(SIOKD), a
		in	a, (SIOKD)
		and	70h
		jr	z, UR6010_5
		ld	a, 30h
		out	(SIOKD), a
		scf
		ret

UR6010_5:	call	UR6010_u2
		in	a, (SIODA)
		cp	0Ah
		jr	z, UR6010_2
		cp	1Bh
		jr	nz, UR6010_6
		ld	a, 7Fh
		jr	UR6010_8
UR6010_6:	cp	80h
		jr	c, UR6010_8
		ld	hl, UR6010_tab
		ld	bc, 3
		cpir
		jr	nz, UR6010_7
		ld	bc, 2
		add	hl, bc
		ld	a, (hl)
		jr	UR6010_8
UR6010_7:	ld	hl, UR6010_b2
		ld	(hl), a
		ld	a, 1
UR6010_8:	or	a
		ret

UR6010_tab:	db  8Ah
		db  82h
		db  87h
;
		db    3
		db  1Ah
		db  1Bh
;
UR6010_u1:	ld	a, 15h
		out	(SIOKD), a
		ld	a, 0EAh
		out	(SIOKD), a
		ret
;
UR6010_u2:	ld	a, 5
		out	(SIOKD), a
		ld	a, 0E8h
		out	(SIOKD), a
		ret

UR6010_b1:	db 	0
UR6010_b2:	db	0


;-----------------------------------------------------------------------------
; UP2 Gertreiber fr PUNCH
;-----------------------------------------------------------------------------
uup:		push	af
		ld	a, 0Ch
		jr	ucrtl1

;-----------------------------------------------------------------------------
; CRT-Gertetreiber fr	LIST
;-----------------------------------------------------------------------------
ucrtl:		push	af
		ld	a, 7Eh
ucrtl1:		ld	(ucrtl4+1), a
		ld	a, 11001111b
		out	(PIO1BC), a	; PIO1B	init Bitmode
		ld	a, 10000000b
		out	(PIO1BC), a	; Bit7 Eingabe,	Bit6-Bit0 Ausgabe
		pop	af
		inc	a
		ret	z
		dec	a
		jr	nz, ucrtl2
		in	a, (PIO1BD)	; PIO1B	lesen
		or	7Fh
		cpl
		ret
ucrtl2:		call	stopkey
		jr	nc, ucrtl3	; Sprung, wenn STOP-Taste nicht	gedrckt
		ld	(15h), a	; LISW,	Schalter fr Hardcopy
		ret
ucrtl3:		in	a, (PIO1BD)	; PIO1B	lesen
		add	a, a
		jr	c, ucrtl2
		ld	a, c
		cp	7Fh
		jr	nz, ucrtl4
		ld	a, 1Bh
ucrtl4:		ld	e, 7Eh
		ld	b, 9
		di
		or	a
		rla
ucrtl5:		out	(PIO1BD), a	; PIO1B	schreiben
		call	ucrtl6		; kurze	Pause, Zeitwert	in E
		rra
		djnz	ucrtl5
		or	1
		out	(PIO1BD), a	; PIO1B	schreiben
		ei
; kurze	Pause, Zeitwert	in E
ucrtl6:		push	de
ucrtl7:		dec	e
		jr	nz, ucrtl7
		pop	de
		ret

_0_572Fh:	db	0
;-----------------------------------------------------------------------------
; UL Gertetreiber fr LIST
;-----------------------------------------------------------------------------
uul:		inc	a
		ret	z
		dec	a
		cpl
		ret	z
		ld	a, 11001111b
		out	(PIO1BC), a
		xor	a
		out	(PIO1BC), a
		ld	a, 01010111b
		out	(CTC1), a
		out	(CTC1), a
		ld	a, c
		or	80h
		cp	0FFh
		jr	nz, uul1
		ld	a, 9Bh
uul1:		out	(PIO1BD), a
		push	ix
		pop	ix
		and	7Fh
		out	(PIO1BD), a
		or	80h
		out	(PIO1BD), a
uul2:		call	stopkey
		ret	c
		in	a, (CTC1)
		cp	57h
		jr	z, uul2
		xor	a
		ret

;-----------------------------------------------------------------------------
;
;-----------------------------------------------------------------------------

;CPA-Source: BIOSDSK.MAC

; Laufwerksauswahl
;=================
; i C: Laufw. (0=A:, 1=B: ...)
; i E: Bit 0 ist LOGIN-Bit von BDOS (=0, wenn LOGIN)

seldsk:		call	dgetpb			; HL auf DPH stellen, IX auf DPB
		ret	z			; Geraet ex. nicht
		ld	a, c
		ld	(ddrive), a		; Laufwerk merken
		bit	0, e			; LOGIN von BDOS?
		ret	nz			; nein

		push	hl			; ^DPH merken
		ld	hl, dbdev
		cp	(hl)			; LOGIN fuer gepuffertes Device?
		jr	nz, seldsk1		; nein
		ld	(hl), 0FFh		; Puffer ist ungueltig
		ld	hl, dbflg
		res	2, (hl)			;kein veraend. Puffer auszug.!

seldsk1:	bit	6, (ix+0Fh)		;Formaterkennung unterdr.?

		IF BIOSVER = 'CPMZ9OK'

		jp	nz, seldsk23		;ja

		ELSEIF BIOSVER <> 'CPMZ9OK'

		jr	z, seldsk1a
		ld	(byte_5D17), a
		ld	l, (ix+0Dh)
		ld	h, (ix+0Eh)
		ld	(word_5D18), hl
		ld	hl, unk_5D16
		call	dsktra
		jp	z, seldsk23
		pop	hl
		ld	hl, 0
		ret
seldsk1a:	equ	$

		ENDIF

		call	dbtrw			;Puffer freimachen und auf Lesen schalten
		ld	hl, dbnb
		ld	(hl), 0FFh		;Puffernummer ist ungueltig, obwohl evtl.
						;alle anderen Pufferparameter stimmen
						;(da nur 128 Bytes vom Puffer belegt sind)
		ld	a, c
		ld	(dfrmdv), a
		ld	(dbdev), a
		xor	a
		ld	(ix+0Fh), a		;ruecksetzen alle Flags
		ld	(ix+13h), a		;keine Weiter-Numm. auf Ruecks.
		ld	(ix+0Dh), a		;keine Systemspuren
		ld	(dfrmtr), a		;Spur 0
		ld	(dbtrk), a
		IF BIOSVER <> 'CPMZ9OK'
		ld	(dbtrk+1), a
		ENDIF
		inc	a
		ld	(dbsec), a		;Sektor 1
		ld	(dbsnb), a		;1 phys. Sektor lesen
; Stellen  Sektor-Translate-Tabelle 1,2,3,...
		push	ix
		ld	b, 1Ah			;max. Laenge der Tabelle
seldsk2:	ld	(ix+19h), a
		inc	ix
		inc	a
		djnz	seldsk2
		pop	ix
; Positionieren auf Spur 0 und analysieren Systemspuren
		call	dsidtr			;beliebigen Sekt.Id lesen
		jr	nz, seldsk6		;Fehler beim Lesen; Systemsp. annehmen
		ld	a, h
		ld	(dbslc), a		;setzen Spurformat fuer Sp. 0
		ld	hl, dbflg
		ld	b, (hl)			;merken Fehlerprotokoll-Bit
		set	4, (hl)			;kein Fehlerprotokoll
		push	hl
		call	dbtran			;Lesen Spur 0, Sektor 1
		pop	hl
		ld	(hl), b			;setzen Standard fuer Fehlerprotokoll-Bit
		jr	nz, seldsk6		;Fehler, Systemspuren annehmen
		push	iy
		ld	iy, (dbdma)
		ld	a, (iy+0)		;Format-Loeschbyte
		cp	0E5h			;leere Diskette/geloeschter Eintrag?
		jr	nz, seldsk3		;nein, weder/noch
		cp	(iy+1)			;auch danach 0e5h?
		jr	nz, seldsk4		;nein, geloeschter Dir-Eintrag
		dec	(ix+0Dh)		;Flag: leere Disk oder leere Systemsp.
		jr	seldsk4			;dpbofs auf 254
;
seldsk3:	cp	40h			;IBM-Format (SIOS-Daten-Diskette)?
		jr	z, seldsk5		;ja, 0 Systemspuren annehmen
		cp	20h			;<=31 ? ("S"YL ist groesser!)
		jr	nc, seldsk5		;nein, kann kein Directory sein; 0 Systemsp.
;Systemlader fuer PC1715 beginnt mit 02 oder 03
		ld	a, (iy+20h)		;bei SCP1715-Systemdiskette dort 00h, bei MicroDos dort 10h
		or	(iy+21h)		;bei SCP1715-Systemdiskette dort 00h, bei CP/A dort Dir-Eintrag (E5 oder Filename)
		jr	z, seldsk5		;beides 00h, kein Directory; 0 Systemsp.
seldsk4:	dec	(ix+0Dh)		;=255, wenn Directory; =254 wenn vorn leer
seldsk5:	pop	iy

seldsk6:
; Analysieren Datenspur

; SS/DS-Analyse
		ld	(ix+12h), 2		;Doppel-Step-Impulse
		ld	a, 3			;auf LOGIN-Spur
		call	dsidtt			;belieb. SektId Vorderseite lesen
		jr	nz, seldsk9		;Fehler, unsicheres Format
		ld	d, h			;d:=Sektorlcode, e:=Spur

		IF BIOSVER <> 'CPMSS4R'
;if dskds
		bit	0, (ix+18h)		;ist es ein DS-Laufwerk?
		jr	z, seldsk7		;nein, SS
		push	de			;merken Sektorlaengencode und Spur
		set	5, (ix+0Fh)
		ld	a, 7			;Rueckseite der Spur, ab der Format bei SS konstant
		call	dsidtt			;beliebigen Sekt.Id Ruecks. lesen
		ld	a, d			;merken side Ruecks.
		ld	l, e			;h:=Sektorlcode, l:=Spur Rueckseite
		pop	de			;wiederherstellen Vorderseite
		jr	nz, seldsk7		;Fehler beim Lesen, SS
		or	a
		sbc	hl, de			;Vorder- gleich Rueckseite?
		jr	z, seldsk8		;ja; 40 Tr/ 80 Tr unterscheiden
seldsk7:	res	5, (ix+0Fh)

		ENDIF

; 40/80 Track Analyse
seldsk8:	ld	a, e			;trk
		sub	3			;waren 2 Steps richtig?
		jr	z, seldsk10		;ja
		dec	(ix+12h)		;Einzelstep-Impulse
		sub	3			;waere 1 Step richtig?
		jr	z, seldsk10		;ja
						;nein, unsicher
seldsk9:	ld	(ix+10h), 0FFh		;provozieren 'BAD SECTOR'
		ld	hl, 0			;erzeugen 'SELECT' Error
		ex	(sp), hl
		jp	seldsk23

; DS/SS und 40/80 ist unterschieden, es fehlt Sektorlaenge
seldsk10:	ld	a, d
		ld	(ix+10h), a		;definieren Sektorlaengencode
		or	a			;Sektorlaenge 128?
		jr	nz, seldsk11		;nein, Puffer reicht nicht

; Stellen Sektor-Translate-Tabelle 1,7,13,..
seldsv:		push	ix
		pop	hl
		ld	bc, 19h
		add	hl, bc
		ex	de, hl
		ld	hl, xlt
		ld	c, 1Ah
		ldir

seldsk11:	call	dtrsla			;stellen hl entspr. Sektorlaenge
		jr	nz, seldsk9		;unzulaessig
		ld	bc, 5			;Laenge eines Eintrags
		bit	1, (ix+12h)		;40 Tr. auf 80er LW?
		jr	nz, seldsk12		;ja
		ld	a, (ix+16h)		;40er LW?
		cp	40+1
		jr	c, seldsk12		;ja, 40 Tr. auf 40er LW
		add	hl, bc			;auf 80er Format stellen
seldsk12:	bit	5, (ix+0Fh)		;DS?
		jr	z, seldsk13		;nein
		add	hl, bc			;auf DS Format
		add	hl, bc

; DPB modifizieren entsprechend erkanntem Format
seldsk13:	ld	c, (hl)			;Zahl der benutzten Spuren
		ld	(ix+14h), c
		inc	hl
		ld	c, (hl)			;Anzahl 128er Sektoren pro Spur
		ld	(ix+0),	c
		inc	hl
		ld	c, (hl)			;Anzahl der Dir-Eintraege -1
		inc	hl
		ld	a, (hl)			;Anzahl der Systemspuren
		srl	a			;feste Anzahl erzwungen?
		jr	c, seldsk15		;ja
		bit	7, (ix+0Dh)		;kann Spur 0 Directory sein?
		jr	z, seldsk15		;nein, Standardanzahl setzen
		inc	(ix+0Dh)		;evtl. leere Systemspur?
		jr	z, seldsk14		;nein, 0 Systemspuren
		ld	b, a			;retten Anzahl Systemspuren
		ld	(dbtrk), a		;einlesen evtl. moeglicher Datenbeginn
		push	hl
		call	dbtran			; dauert bei falscher Spur etwas laenger
		pop	hl
		ld	a, b
		jr	nz, seldsk15		;Fehler, mit Systemspuren annehmen
		push	iy
		ld	iy, (dbdma)
		ld	a, (iy+0Eh)		;falls dort Dir, so Byte 14 von Dir =0
		cp	0E5h			;Daten auch leer?
		pop	iy
		ld	a, b
		jr	nz, seldsk15		;nein, vorn liegen nichtbelegte Systemspuren
seldsk14:	xor	a			;sonst ohne Systemspuren
seldsk15:	ld	(ix+0Dh), a
		or	a			;0 Systemspuren?
		jr	z, seldsk16		;ja
		ld	a, c			;Anzahl Dir-Eintraege -1
		cp	192-1			;>=192 Dir-Eintraege ?
		jr	c, seldsk16		;nein
		ld	c, 128-1		;780k hat 128 Dir-Eintraege
seldsk16:	ld	(ix+7),	c
		inc	hl
		ld	c, (hl)			;Abstand der Blockgroessentab.
		ld	b, 0
		add	hl, bc			;auf Blockgroessentabelle
		ld	c, (hl)
		ld	(ix+2),	c
		inc	hl
		ld	c, (hl)
		ld	(ix+3),	c
		inc	hl
		ld	c, (hl)
		ld	(ix+4),	c

; Bestimmen Zahl der abweichenden 128er-Sektoren ab Spur 0
		xor	a
		ld	(dfrmtr), a		;ab Spur 0
						;(gleichz. Spurkorr. bei falschen Doppelstepimpulsen)
seldsk17:	call	dsidtr			;beliebigen SektId lesen
		jr	nz, seldsk18		;Fehler
		ld	a, h			;Sektorlaenge
		cp	(ix+10h)		;gleich Disketten-Rest?
		jr	z, seldsk18		;ja
		ld	a, (ix+0Fh)
		and	3			;bisherige Anzahl abweichender Spuren
		cp	3			;Zaehler voll?
		jr	z, seldsk18		;ja
		inc	(ix+0Fh)		;erhoehen abweich. Spurzahl
		ld	hl, dfrmtr		;naechste (logische!) Spur
		inc	(hl)
		jr	seldsk17

seldsk18:
; Setzen Anzahl der logischen 128er Rekords im Puffer -1
; Es wird immer das Maximum gesetzt, dies setzt eine dichte
; Sektorfolge beim Transfer voraus! (so in dpbstr)
		ld	(ix+11h), 7

; Berechnen Speicherkapazitaet-1 in BDOS-Bloecken aus
; ((Tracks-dpbofs)*dpbspt/(2**dpbbls))-1
		ld	l, (ix+14h)
		xor	a
		ld	h, a			;hl:=log. Spurzahl
		ld	e, (ix+0Dh)
		ld	d, a			;de:=log. offset-Spurzahl
		sbc	hl, de
		ex	de, hl			;de:=log. Daten-Spurzahl
		ld	l, a			;hl:=0
		ld	b, (ix+0)
seldsk19:	add	hl, de
		djnz	seldsk19		;hl:=(tracks-dpbofs)*dpbspt
		ld	b, (ix+2)
seldsk20:	srl	h
		rr	l
		djnz	seldsk20
		dec	hl
		ld	(ix+5),	l
		ld	(ix+6),	h
		ld	a, (ix+7)		;Dir-Groesse
		inc	a
		IF  BIOSVER <> 'CPMZ9OK'
		and	0FCh
		ENDIF
		rrca
		rrca				;div 4, da 4 Dir-Eintr. /Sekt.
		ld	(ix+0Bh), a		;ist Sektorzahl = Check size
		ld	b, (ix+2)		;block shift
seldsk21:	rrca				;BDOS-Bloecke fuer Dir.
		djnz	seldsk21
		ld	b, a
		xor	a
seldsk22:	scf				;Allocation-Bits fuer Dir.
		rra
		djnz	seldsk22
		ld	(ix+9),	a

seldsk23:	pop	hl			;hl auf DPH
		ret				;Return Seldsk

;-----------------------------------------------------------------------------
; home
; auf Spur 0 zurueck (vor jedem Dir-Zugriff)
;-----------------------------------------------------------------------------

home:		ld	hl, dbflg
		bit	2, (hl)			;veraenderter Puffer?
		jr	nz, home1		;ja, nicht freigeben
		ld	a, 0FFh			;sonst Diskwechsel erlauben
		ld	(dbdev), a		;Puffer ist nicht aktiv

;-----------------------------------------------------------------------------
; settrk
; Einstellen Spur in Reg. BC
;-----------------------------------------------------------------------------
		IF BIOSVER = 'CPMZ9OK'

home1:		ld	c, 0
settrk:		ld	a, c
		ld	(dtrack), a		;Spur merken
		ret

		ELSEIF BIOSVER <> 'CPMZ9OK'

home1:		ld	bc, 0
settrk:		ld	(dtrack), bc
		ret

		ENDIF


;-----------------------------------------------------------------------------
; Einstellen Sektor in Reg. C
;-----------------------------------------------------------------------------
setsec:		ld	a, c
		ld	(dsectr), a
		ret

;-----------------------------------------------------------------------------
; Einstellen DMA in Reg. BC
;-----------------------------------------------------------------------------
setdma:		ld	(ddma), bc
		ret

;-----------------------------------------------------------------------------
; Uebersetzung Sektornummer
; Translate-Tab-Adr. in DE, Eingangs-Sektornummer in BC,
;			    Ausgangs-Sektornummer in HL
; Es wird keine Translate-Tabelle benutzt, da die Sektor-
; nummernverwaltung verallgemeinert im nicht-Standard-DPB
; enthalten ist (auch fuer physische Sektorlaenge <>128)
;-----------------------------------------------------------------------------
sectran:	ld	h, b
		ld	l, c
		inc	hl		;Sektoren zaehlen in CP/A ab 1
		ret

; Ermitteln DPH und DPB von (c) nach HL bzw. IX
; ret z		DPH existiert nicht, HL=0
; ret nz	ok, HL auf DPH, IX auf DPB
; BC,DE bleibt erhalten
dgetpb:		push	de
		ld	hl, 0
		ld	a, c
		cp	2		; nur LW 0 und 1 erlaubt
		jr	nc, dgetpb1 	; sonst Abbruch
		ex	de, hl		; DE = 0
		ld	hl, dphtab	; Adresstabelle fuer DPH's
		ld	e, c
		add	hl, de
		add	hl, de
		ld	a, (hl)
		inc	hl
		ld	h, (hl)
		ld	l, a		; HL=Adr. zum LW gehriger DPH
		push	hl
		ld	e, 10
		add	hl, de
		ld	e, (hl)
		inc	hl
		ld	d, (hl)		; de:=^dpb
		push	de
		pop	ix		; IX=Adr. zugehriger DPB
		pop	hl
dgetpb1:	pop	de
		ld	a, h
		or	l		; ret z bei hl=0
		ret

;-----------------------------------------------------------------------------
; Schreiben Sektor
; Register C (vom BDOS gestellt):
;	=0, wenn normales write
;	=1, wenn directory-write (sofort ausgeben)
;	=2, wenn Beginn eines neuen Datenblocks (kein preread)
;-----------------------------------------------------------------------------
write:		ld	e, c			;retten Reg. C
		ld	a, (ddrive)
		ld	c, a
		call	dgetpb			;ix:=dpb(ddrive)
		jr	z, drwerr		;Geraet ex. nicht

		ld	hl, dflg
		set	2, (hl)			;Write-Flag setzen
		ld	hl, dbflg
		set	0, (hl)			;Annahme preread notwendig
		ld	a, e			;A:= Write-Typ
		ld	(dwrtyp), a		;merken Write-Typ
		cp	2			;write to unallocated?
		jr	nz, write1		;nein
		ld	a, (ix+3)		;Zahl der 128-Sekt. im Block -1
		inc	a
		ld	(unacnt), a		;fuer diese Sekt. kein preread
		ld	hl, unadev
		ld	(hl), c

		IF BIOSVER = 'CPMZ9OK'
		ld	a, (dtrack)
		ld	(unatrk), a
		ld	a, (dsectr)
		ld	(unasec), a
		ELSEIF BIOSVER <> 'CPMZ9OK'
		ld	hl, (dtrack)
		ld	(unatrk), hl
		ld	a, (dsectr)
		ld	(dbnb), a
		ENDIF

write1:		ld	hl, unacnt
		ld	a, (hl)
		or	a			;noch nicht geschr. Sekt. da?
		jr	z, alloc		;nein
		dec	a			;sonst Restzahl -1
		ld	(hl), a
		inc	hl
		ld	a, c
		cp	(hl)			;gleiches Geraet?
		jr	nz, alloc		;nein
		inc	hl
		ld	a, (dtrack)
		cp	(hl)			;gleiche Spur?
		jr	nz, alloc		;nein

		IF BIOSVER <> 'CPMZ9OK'

		inc	hl
		ld	a, (dtrack+1)
		cp	(hl)
		jr	nz, alloc

		ENDIF

		inc	hl
		ld	a, (dsectr)
		cp	(hl)			;gleicher Sektor?
		jr	nz, alloc		;nein
; Vorbereiten naechsten unalloc write-Aufruf
		cp	(ix+0)			;neue Spur?
		jr	c, write2		;nein

		IF BIOSVER = 'CPMZ9OK'

		dec	hl
		inc	(hl)			;unatrk +1
		inc	hl

		ELSEIF BIOSVER <> 'CPMZ9OK'

		dec	hl
		ld	d, (hl)
		dec	hl
		ld	e, (hl)
		inc	de
		ld	(hl), e
		inc	hl
		ld	(hl), d
		inc	hl

		ENDIF

		xor	a			;Sektor auf Spuranfang
write2:		inc	a			;naechster Sektor
		ld	(hl), a
		ld	hl, dbflg
		res	0, (hl)			;anzeigen kein preread notw.
		jr	drw

; Abbruch read/write mit E/A-Fehler
drwerr:		ld	a, 1
		ret

;-----------------------------------------------------------------------------
; Lesen Sektor
;-----------------------------------------------------------------------------
read:		ld	a, (ddrive)
		ld	c, a
		call	dgetpb			;ix:=dpb(ddrive)
		jr	z, drwerr		;Geraet ex. nicht
		ld	hl, dflg
		res	2, (hl)			;Lesen anzeigen
		ld	hl, dbflg
		set	0, (hl)			;preread notwendig
		ld	a, 2
		ld	(dwrtyp), a

alloc:		xor	a
		ld	(unacnt), a		;Ende unalloc

; gemeinsamer Zweig read/write Floppy
drw:		ld	a, (dsectr)		;Sektornummer in 1..dpbspt ?
		dec	a			;ab 0 zaehlen
		jp	m, drwerr		;<0, Fehler
		cp	(ix+0)			;<dpbspt?
		jr	nc, drwerr		;nein, Fehler
		ld	b, a			;retten dsectr-1
		ld	a, (ix+0Fh)
		and	3			;Anzahl der abweich. 128er Spuren

		IF BIOSVER = 'CPMZ9OK'

		ld	d, a
		ld	a, (dtrack)		;verlangte Spur
		cp	d			;im abweichenden Spurformat?

		ELSEIF BIOSVER <> 'CPMZ9OK'

		ld	e, a
		ld	d, 0
		ld	hl, (dtrack)
		or	a
		sbc	hl, de

		ENDIF

		jp	c, drwsec		;ja, Sektorlaenge 128

		ld	a, (ix+10h)		;phys. Sektorlaenge 128?
		or	a			;d.h. mit Sektorversatz?
		jp	z, drwsec		;ja, keine Pufferung sinnvoll

		ld	a, b
		ld	d, (ix+11h)		;Puffermaske
		and	d			;rel. Sekt.nr. des BDOS-Blocks zum Puff.anf.
		push	af			;merken fuer move
		xor	b			;a ist Sektornr. des Puff.anf.
drwbnr:		srl	a			;ermitteln Puffernr in Spur
		rr	d			;noch Bits in Puffermaske?
		jr	nz, drwbnr		;ja
; Test, ob dieser Puffer da ist
		push	af			;merken Puffernummer
		ld	hl, dbnb
		cp	(hl)			;richtige Puffernummer?
		jr	nz, dbn			;nein

		IF BIOSVER = 'CPMZ9OK'
		ld	hl, dbtrk
		ld	a, (dtrack)		;geforderte Spur
		cp	(hl)			;richtige Spur?
		ELSEIF BIOSVER <> 'CPMZ9OK'
		ld	hl, (dtrack)
		ld	de, (dbtrk)
		or	a
		sbc	hl, de
		ENDIF

		jr	nz, dbn
		ld	a, (dbdev)		;zum Puffer gehoeriges Geraet
		cp	c			;richtiges Geraet?
		jr	z, dbmat		;ja, Sektor steht im Puffer

dbn:		push	bc			;retten ddrive
		call	dbtrw			;veraenderten Puffer ausgeben
		pop	bc			;c:=ddrive
		ld	hl, dbdev
		ld	(hl), c

		IF BIOSVER = 'CPMZ9OK'
		ld	a, (dtrack)
		ld	(dbtrk), a
		ELSEIF BIOSVER <> 'CPMZ9OK'
		ld	hl, (dtrack)
		ld	(dbtrk), hl
		ENDIF

		call	dgetpb			;ix:=dpb(c)
		ld	b, (ix+10h)
		ld	hl, dbslc
		ld	(hl), b
		ld	a, (ix+11h)		;Zahl der 128er Sektoren -1
		inc	a
		add	a, a			;CY:=0
		inc	b			;fuer djnz, falls dpbslc=0
dbsnbz:		rrca				;Zahl der phys. Sekt. im Puff.
		djnz	dbsnbz
		ld	d, a			;merken
		xor	a
		pop	bc			;b:=geforderte Puffernr.
		push	bc
		ld	hl, dbnb
		ld	(hl), b
		inc	b
		dec	b			;Puffernr. 0?
		jr	z, dbsec0		;ja
dbsecz:		add	a, d			;0. phys. Sekt. im Puffer
		djnz	dbsecz			;=dbsnb.*Puffernr.
dbsec0:		ld	e, a			;merken Pufferanf.-Sektor
		ld	b, (ix+10h)
		ld	a, (ix+0)
		inc	b			;wegen djnz
		add	a, a
dbsecm:		srl	a			;phys. Sektoranzahl
		djnz	dbsecm
		sub	e			;Restsektorzahl auf Spur
		cp	d			;>= Zahl zu transferierender?
		jr	nc, dbseco		;ja
		ld	d, a			;sonst nur Restsektorzahl
dbseco:		inc	e			;Sektoren zaehlen ab 1
		ld	a, e
		ld	(dbsec), a		;Sektor
		ld	a, d
		ld	(dbsnb), a		;Sektorzahl
		ld	hl, dbflg
		res	2, (hl)
		bit	0, (hl)			;muss neuer Puffer gel. werden?
		jr	nz, dbprr		;ja
; nur dann kein preread, wenn Rest BDOS-Block >= phys. Puff.
; und Rest BDOS-Block auf Pufferanfang beginnt
		ld	a, (unacnt)
		cp	(ix+11h)		;Restzahl-1 >= Puffergr.-1 ?
		jr	c, dbprr		;nein, preread
		ld	a, (dsectr)		;Sektornr.
		dec	a			;ab 0 zaehlen
		and	(ix+11h)		;rel. Sekt.nr. des BDOS-Blocks zum Pufferanfang
		jr	nz, dbprr
		ld	(dberrf), a		;fehlerfrei gel. anzeigen
dbprr:		call	nz, dbtran		;Puffer lesen

; Sektor in/aus Puffer holen
dbmat:		pop	af			;Puffernr. in Spur wegschm.
		pop	af			;128-er Index im Puffer
		push	af
		rra				;Sektornb*128 berechnen
		ld	b, a
		ld	c, 0
		rr	c
		ld	hl, fdcbuffer
		add	hl, bc			;Adresse im Puffer
		ld	de, rwbuffer	;--> in CPA: ld	de,(ddma)	;Nutzer-DMA
		ld	a, (dflg)
		bit	2, a			;Schreiben?
		jr	z, dsmove		;nein
		ld	a, (dbflg)
		set	2, a			;anzeigen Puffer beschrieben
		ld	(dbflg), a
		ex	de, hl
dsmove:		ld	bc, 128
		call	flldir
		pop	af			;128-Index im Puffer
		cp	(ix+11h)		;letzter Sektor im Puffer
		jr	z, dmovew		;ja, Puffer ausgeben
		ld	a, (dwrtyp)
		cp	1			;write to directory?
dmovew:		call	z, dbtrw		;ja, veraend. Puffer ausgeben
		ld	a, (dberrf)		;Ergebnis letztes dbtran
		ret

; Schreiben Puffer, wenn notw.
dbtrw:		ld	hl, dbflg
		bit	2, (hl)			;Puffer veraendert?
		ret	z			;nein, Schreiben unterdruecken

; gemeinsamer Zweig Lesen/Schreiben Puffer
dbtran:		ld	hl, dbflg
		call	dsktra
		ld	(dberrf), a		;Fehlerflag stellen
		ld	hl, dbflg
		res	2, (hl)			;Puffer ist nicht beschrieben
		ret

; Lesen beliebigen Sektor-Id.
dsidtt:		ld	(dfrmtr), a		;Eingang fuer Spur in A
dsidtr:		ld	hl, dfrcdb
		jr	dsktra

; gemeinsamer Zweig Read/Write 128-Sektor einzeln
drwsec:		ld	hl, dflg


;*********************************************************
; Umkleidung des physischen Disketten-Transfers
;*********************************************************

; Parameter: HL auf CDB mit folgender Struktur:
; +0: cdbfl	;Flags	(Struktur angepasst an ft.kom)
diof00	equ	0	;** frei fuer Anw. ** (frueher Verify nach Schreiben)
diofid	equ	1	;fest	;=1, wenn nur Sektid. zu lesen
diofwr	equ	2	;fest	;=1, wenn Schreiben
dioffm	equ	3	;fest	;=1, wenn temporaer FM-Format erzwungen
dioftr	equ	4		;=1, wenn keine Fehlermeldung (und -behandlg)
diofhd	equ	5		;=1, wenn Kopf hochnehmen ("noerr")
diofps	equ	6		;=1, wenn trk,sid,sec schon physisch
diofs1	equ	7	;fest	;=1, wenn Rueckseite

; ab hier unwichtig bei "diofhd" in cdbfl =1 (noerr)
; +1: cdbdev	;logisches Geraet 0 .. dphnb-1
; +2: cdbtrk	;log./phys. Spur
; +3: cdbsid	;bel./phys. Seite
; +4: cdbsec	;log./phys. Sektor
; +5: cdbslc	;Sektorlaengencode (0=128, 1=256, 2=512, 3=1024)
; +6: cdbsnb	;Anz. zu uebertr. ph. Sekt. (wenn =0, so nur position.)
; ab hier nur wichtig, wenn "diofid" in cdbfl =0:
; +7,8: cdbdma	;Transferadresse

; Return, falls nicht "diofhd" in cdbfl =1:
;	A=0 (ret z) bei fehlerfrei, sonst A=1 (ret nz) bei cdbfl, "dioftr" =0
;	oder Returncode phys. Transfer
;	E:=trk, D:=sid ,L:=sec, H:=len
;	BC,IX unveraendert

dsktra:		push	bc
		ld	de, diocdb		;CDB des Nutzers auf internen Hilfsspeicher
		ld	bc, 9			;da modifiziert
		push	de
		call	flldir			;falls Original im ROM-Adressbereich
		pop	hl			;hl auf diocdb
		pop	bc
		bit	5, (hl)			;Kopf hochnehmen?
		jp	nz, noerr		;ja, kein Transfer
		push	bc
		push	ix
		push	hl
		inc	hl
		ld	c, (hl)			;logisches Laufwerk
		call	dgetpb			;IX auf DPB stellen
		pop	hl			;wiederherstellen hl
		ld	a, 'D'			;falls LW nicht ex.
		jp	z, dioert		;LW ex. nicht
		bit	6, (hl)			;trk,sid,sec schon physisch?
		jr	nz, diophy		;ja
; Track, Side und Sector entpsr. Diskettenformat setzen
		inc	hl
		inc	hl
		ld	d, (hl)			;Track
		ld	e, d			;merken
		inc	hl
		inc	hl
		ld	c, (hl)			;Sektornummer ab 1
		dec	c			;Sektornr. ab 0
		bit	4, (ix+0Fh)		;Fortsetzung Dsk. auf Ruecks.?
		jr	nz, diots2		;ja
		bit	5, (ix+0Fh)		;ungerade Spuren auf Rueckseite?
		jr	z, diotrs		;nein, einseitig
		srl	d			;Spur halbieren
		jr	nc, diotrs		;gerade Spur, Vorderseite
		jr	diots1			;auf Rueckseite
diots2:		ld	b, (ix+14h)		;log. Spurzahl
		srl	b			;Spuren auf Vorderseite
		ld	a, d
		sub	b			;Spur auf Vorderseite?
		jr	c, diotrs		;ja, Spur in d unveraendert lassen
						;40 ->  0; 41 ->  1; ...; 79 ->  39
						;77 ->  0; 78 ->  1; ...;153 ->  76
		bit	3, (ix+0Fh)		;Ruecks. von aussen nach innen?
		jr	nz, diots4		;ja
		cpl				;40 -> -1; 41 -> -2; ...; 79 -> -40
						;77 -> -1; 78 -> -2; ...;153 -> -77
		add	a, b			;40 -> 39; 41 -> 38; ...; 79 ->   0
						;77 -> 76; 78 -> 75; ...;153 ->   0
diots4:		ld	d, a
diots1:		ld	a, (diocdb)
		set	7, a			;vermerken Rueckseite
		ld	(diocdb), a
		ld	a, (ix+13h)		;Versch. der Sektornr. auf Rueckseite
		add	a, c			;Sektornr. evtl. weiterzaehlen
		ld	c, a
		ld	a, 1			;side:=1
		jr	diotss

diotrs:		xor	a			;side:=0
diotss:		ld	(diocsi), a		;phys. Seite setzen
		ld	a, d
		ld	(dioctr), a		;phys. Spur setzen
		ld	a, e			;log. Spurnr. zurueck nach A
		push	ix
		pop	hl
		ld	de, 19h			;d:=0
		add	hl, de
		cp	(ix+0Dh)		;Systemspur?
		jr	c, diossp		;ja, Sektornummern nicht logisch
		ld	e, c
		add	hl, de			;HL entspr. Index in dpbstr
		ld	c, d			;c:=0
diossp:		ld	a, c			;phys. Sektornr. ab 0
		add	a, (hl)			;+echte Sektornr.
		ld	(diocse), a		;phys. Sektor setzen
diophy:
; i ret, BC, IX im Stack
; i cdb auf diocdb
; i IX auf DPB

;interne Wiederholung dsktra
		push	ix			;retten IX fuer evtl. Wiederholung
		ld	hl, diocdb
		push	hl
		ld	a, (hl)			;side 0/1, -/FM, read/write, -/sectid l.
		and	8Eh			;(1 shl diofs1)+(1 shl dioffm)+(1 shl diofwr)+(1 shl diofid)
		ld	b, a
		ld	a, (ix+18h)		;5"/8", FM/MFM, 40/80, Verify nach Schreiben
		and	78h			;and (1 shl dpbt80)+(1 shl dpbt5z)+(1 shl dpbtdd)+(1 shl dpbtwv) ;aus DPB
		xor	b
		bit	4, (hl)			;Fehlerbehandlung unterdruecken?
		jr	z, dionft		;nein
		inc	a
dionft:		ld	(ft.kom), a
		inc	hl
		ld	a, (ix+15h)		;physische LW-Nr.
		ld	(ft.lwn), a
		inc	hl
		ld	a, (hl)
		ld	(ft.trk), a
		inc	hl
		ld	a, (hl)
		ld	(ft.sid), a
		inc	hl
		ld	a, (hl)
		ld	(ft.sec), a
		inc	hl
		ld	a, (hl)
		ld	(ft.len), a
		inc	hl
		ld	a, (hl)
		ld	(ft.anz), a		;Anzahl der Stepimpulse
		ld	a, (ix+12h)		;Schrittzeit
		ld	(ft.stp), a
		ld	a, (ix+17h)
		ld	(ft.sti), a
		ld	hl, (diocad)
		ld	(ft.adr), hl
		call	floppy			;$$$ phys. Transfer $$$
		pop	hl			;HL wieder auf Paramfeld-Adresse
		pop	ix			;IX wieder auf DPB

; Fehlerprotokoll
dioert:		bit	4, (hl)			;Fehlerprotokoll und -behandlung  unterdr.?
		jr	nz, dtrok		;ja
		or	a			;Returncode
		jr	z, dtrok		; -> fehlerfrei
		cp	'R'			;not ready?
		jp	z, diophy		;ja, Wiederholung
		ld	(derrcd), a		;Fehlercode
		bit	2, (hl)			;war lesen?
		ld	a, 'R'
		jr	z, derrr		;ja
		ld	a, 'W'
derrr:		ld	(derrrw), a
		ld	hl, ft.trk		;phys. Spur
		ld	de, derrtr	; "XXXXXX\a"
		call	mbreco		; (HL) nach hex	konvertieren, Eintragen	nach (DE), 2x inc DE
		ld	hl, ft.trk+1
		call	mbreco		; (HL) nach hex	konvertieren, Eintragen	nach (DE), 2x inc DE
		ld	hl, ft.sec		;Seite
		call	mbreco		; (HL) nach hex	konvertieren, Eintragen	nach (DE), 2x inc DE
		ld	hl, derrms
		call	biosms
		ld	a, 1		;return mit Fehler
dtrok:		or	a		;stellen Z-Flag
		ld	de, (ft.trk)	;E:=trk, D:=sid
		ld	hl, (ft.sec)	;L:=sec, H:=len
		pop	ix
		pop	bc
		ret

; Stellen HL entsprechend Sektorlaengencode in (A)
; ret nz bei unzulaessigem
dtrsla:		or	a
		ld	hl, dtrsl0	; bei A=0 (128)
		ret	z
		dec	a
		ld	hl, dtrsl1	; bei A=1 (256)
		ret	z
		dec	a
		ld	hl, dtrsl2	; bei A=2 (512)
		ret	z
		dec	a
		ld	hl, dtrsl3	; bei A=3 (1024)
		ret

;************************************************
;	Steuertabellen
;************************************************

; Modifizierungstabellen entspr. Sektorlaenge und LW-Typ
;-------------------------------------------------------

; Struktur:

dsltrk	equ	0		;benutzte Spuren
dslspt	equ	dsltrk+1	;Sektoren/Spur
dsldir	equ	dslspt+1	;Dir-Eintraege
dsloff	equ	dsldir+1	;2*offset
				;[+Flag fuer festes offset]
dslfo	equ	1		;festes offset
dslvo	equ	0		;offset =0, falls Directory mgl.
dslblk	equ	dsloff+1	;rel. Adr. Tab. BDOS-Blockgroesse
dsll	equ	5		;Laenge eines Eintrags

; Reihenfolge
; 5", 40 Tr, SS
; 5", 80 Tr, SS
; 5", 40 Tr, DS
; 5", 80 Tr, DS

; 128
dtrsl0:
	db	40,26,63,2*2+dslvo,dbl1k-$-4	;CP/M Standard
	db	80,26,127,2*2+dslvo,dbl2k-$-4
	db	80,26,127,2*0+dslfo,dbl2k-$-4
	db	160,26,127,2*0+dslfo,dbl2k0-$-4

; 256
dtrsl1:
	db	40,32,63,2*3+dslfo,dbl2k-$-4	;SCP Hausformat A51xx
	db	80,32,63,2*3+dslfo,dbl2k-$-4	;SCP
	db	80,32,127,2*4+dslfo,dbl2k-$-4
	db	160,32,127,2*4+dslfo,dbl2k0-$-4	;SCP Hausformat PC1715

; 512
dtrsl2:
		IF BIOSVER = 'CPMZ9OK'

	db	40,36,63,2*0+dslfo,dbl1k-$-4
	db	80,36,127,2*0+dslfo,dbl2k-$-4
	db	80,36,127,2*0+dslfo,dbl2k-$-4
	db	160,36,127,2*0+dslfo,dbl2k0-$-4

		ELSEIF BIOSVER <> 'CPMZ9OK'

	db	40,36,0,2*0+dslfo,dbl1k-$-4
	db	80,36,0,2*0+dslfo,dbl2k-$-4
	db	80,36,0,2*0+dslfo,dbl2k-$-4
	db	160,36,0,2*0+dslfo,dbl2k0-$-4

		ENDIF

; 1024
dtrsl3:
	db	40,40,63,2*2+dslvo,dbl1k-$-4	;CP/A Standard A51xx
	db	80,40,127,2*2+dslvo,dbl2k-$-4	;CP/A
	db	80,40,127,2*0+dslvo,dbl2k-$-4	;CP/A
	db	160,40,191,2*4+dslvo,dbl2k0-$-4	;CP/A (und SCP)

; Modifizierungstabellen fuer BDOS-Blockgroesse
dbl1k:	db	3,7,0		;Bl. Shift, Bl. Mask, Ext. Mask
dbl2k:	db	4,0fh,1
dbl2k0:	db	4,0fh,0		;fuer Kapazitaet >255

; log. Sektortranslate-Tabelle fuer 26*128
xlt:	db	1,7,13,19,25,5,11,17,23,3,9,15,21
	db	2,8,14,20,26,6,12,18,24,4,10,16,22


; Fehlermeldung Disk I/O Error
; Genau so lang (kurz), wie in Statuszeile Platz dafuer da ist!
; und das sind 17 Bytes!

derrms:		db	 0Dh ;
		db	 0Ah ;
derrrw:		db	'X'
derrcd:		db	"X;T,Si,Se="
derrtr:		db	"XX"
derrts:		db	"XX"
derrsc:		db	"XX"
		db	7
		db	0

		IF BIOSVER <> 'CPMZ9OK'
unk_5D16:	db  10h
byte_5D17:	db 0FFh
word_5D18:	dw 0
		db    1 ;
		db    1 ;
		db    0 ;
		ENDIF

; CDB fuer Bestimmung Spurformat
dfrcdb:		db	 12h 		;(1 shl diofid)+(1 shl dioftr)
dfrmdv:		db	0FFh		;Geraet
dfrmtr:		db	0FFh		;Spur
		db	0FFh 		;Seite (bel.)
		db	   1 		;Sektor (bel.)
		IF BIOSVER = 'CPMZ9OK'
		db	   0 		;Sektorlaengencode (bel.)
		ELSEIF BIOSVER <> 'CPMZ9OK'
		db	   1 		;
		ENDIF
		db	   1 		;Sektoranzahl (>0)

; Puffer-CDB
dbflg:	db	0
dbdev:	db	0FFh
dbtrk:	db	0FFh
dbsid:	db	0FFh
dbsec:	db	0FFh
dbslc:	db	0FFh
dbsnb:	db	0FFh
dbdma:	dw	fdcbuffer

; CDB fuer ungepufferte E/A
dflg:	db	0
ddrive:	db	0
dtrack:	db	0FFh
	db	0FFh 		;Seite
dsectr:	db	0FFh
	db	   0 		;128 Bytes Sektorlaenge
	db	   1 		;immer nur 1 Sektor

unacnt:	db	0
unadev:	db	0FFh
unatrk:	db	0FFh
unasec:	db	0

dbnb:	db	0FFh
dwrtyp:	db	0FFh
dberrf:	db	0		;Ergebnisflag letztes dbtran

;Arbeitsbereiche logischer Floppy-Treiber im BIOS-RAM
flldir:		ldir		;Floppy-Puffer <-> Nutzer-DMA
		ret

;CDB-Hilfsspeicher fuer dsktra-Aufruf
diocdb:	db	0FFh
diocde:	db	0FFh
dioctr:	db	0FFh
diocsi:	db	0FFh
diocse:	db	0FFh
diocsl:	db	0FFh
diocsn:	db	0FFh
diocad:	dw	0FFFFh


;************************************************
; physischer Disketten-Transfer
;************************************************

	include biosfdc.inc


	IF BIOSVER = 'CPMZ6010'

; ???????? alte Reste ??????

                db  45h ; E
                db  44h ; D
                db  11h
                db  47h ; G

           	ld      b, h
                ld      c, 1Ah
                call    5
                ld      hl, 40D5h
                ld      a, (hl)
                or      a
                jp      m, 4B7Fh
                jp      4B73h
                push    hl
                push    de
                ld      de, 0Ch
                add     hl, de
                xor     a
                ld      d, 4
                ld      (hl), a
                inc     hl
                dec     d
                jp      nz, 4BAAh
                ld      de, 10h
                add     hl, de
                nop
                nop

	ENDIF


;-----------------------------------------------------------------------------
; Ausgabe Zeichenkette
;-----------------------------------------------------------------------------
prnst:		ex	de, hl
		ld	hl, (0F057h)	; Adr. PRNST-Funktion des Z1013-OS
		jp	(hl)
biosms:		di
		call	prnst
		ei
		ret

;-----------------------------------------------------------------------------
; Kurze Warteschleife
;-----------------------------------------------------------------------------
wait:		push	bc
		ld	c, 0Ah
wait1:		ld	b, 0BDh
wait2:		djnz	wait2
		dec	c
		jr	nz, wait1
		pop	bc
		djnz	wait
		ret

;-----------------------------------------------------------------------------
; (HL) nach hex	konvertieren, Eintragen	nach (DE), 2x inc DE
;-----------------------------------------------------------------------------
mbreco:		ld	a, (hl)
hexa:		call	hexa1		; A nach hex konvertieren, Eintragen nach (DE),	2x inc DE
hexa1:		rrca
		rrca
		rrca
		rrca
		push	af
		and	0Fh
		sub	0Ah
		jr	c, hexa2
		add	a, 7
hexa2:		add	a, 3Ah
		ld	(de), a
		inc	de
		pop	af
		ret

;-----------------------------------------------------------------------------
; wboot,
;-----------------------------------------------------------------------------

wboot:		ld	sp, 80h
		ld	hl, SHADOWRAM	; CCP restaurieren
		ld	de, CCP
		ld	bc, 800h	; Lnge CCP
		ldir

		IF BIOSVER <> 'CPMZ9OK'
		ld	hl, CCP+0003h
wboot3:		ld	(wboot3u1+1), hl
		ENDIF

		di
		ld	a, hi(intvectab)	;E7H
		ld	i, a
		im	2

		ld	a, 0C3h		; Befehlscode JP
		ld	(0), a		; 0000h: jp wboot
		ld	(38h), a	; 0038h: RST 38H, jp wboot
		ld	hl, ubios+3	; wboot	im upper Bios
		ld	(1), hl
		ld	(39h), hl
		ld	(5), a		; 0005h: jp bdos
wboot0:		ld	hl, BDOS
		ld	(6), hl

		ld	a, 0C7h
		out	(CTC3), a
		ld	a, 40h
		out	(CTC3), a

		ld	a, 0FFh
		ld	(dbdev), a

wboot1:		ld	hl, 3
		ld	c, (hl)
		push	bc
		ld	a, c
		and	0Fh
		ld	c, a
		xor	(hl)
wboot2:		or	0
		ld	(hl), a
		ld	e, 1
		call	seldsk
		pop	bc
		ld	a, h
		or	l
		jr	z, wboot1
		jp	wboot3u		; wboot ende und Start CCP


		IF BIOSVER <> 'CPMZ9OK'

; Copyrightschutztest, s. a. copyprot
schutzbyte1:	db	 28h 		; Lo(Adr. zweite Bildschirmzeile)
schutzbyte2:	db	   0 		; Prfwert
schutzbyte3:	dw 	bootmsg		; Adr. Bootmessage
schutzbyte4:	db	0ECh 		; Hi(Adr. zweite Bildschirmzeile)

		ENDIF

;-----------------------------------------------------------------------------
; Aufruf bei call sbios: (jp start), wenn BIOS bereits initialisiert
; die Resert-Routine wird nach 0080h kopiert und gestartet
;-----------------------------------------------------------------------------
ustart:		ld	hl, reset0
		ld	de, 80h
		ld	bc, start-reset0
		ldir
		jp	80h

;-----------------------------------------------------------------------------
; wird nach 0080h kopiert
;-----------------------------------------------------------------------------
reset0:		equ	$
		phase 	0080h
		di
		out	(rambank1), a	; Shadow-Ram deaktivieren
		push	af
		ld	a, 'U'
		ld	(bmode), a
		pop	af
		out	(port71h), a
		ld	(0F800h), a	; ROM-Modul zuschalten

		jp	0F000h		; Sprung in Z9001-OS (Kaltstart)
		dephase



fdcbuffer:	equ	$		; Buffer fr FDC-IO-Operation (1K)

;-----------------------------------------------------------------------------
; boot, vgl. Code mit Initialisierung am Anfang
; Einsprung durch Init nach Bootloader
;-----------------------------------------------------------------------------


start:		di
		ld	hl, LOADADRESS	; anfngliche Ladeadr.
		ld	bc, romend-SHADOWRAM	; Programmlnge
		ld	de, SHADOWRAM
		di
		out	(rambank2), a	; ShadowRAM aktivieren
		push	af
		xor	a
		ld	(bmode), a
		pop	af
		ldir			; BIOS in Shadowram kopieren
		jp	start1		; weiter nach Kopieren des BIOS

;-----------------------------------------------------------------------------
; weiter gehts im Shadow-RAM
; Initialisierung von binit und ubios
;-----------------------------------------------------------------------------

start1:		di
		ld	sp, 80h
		; CCP und BDOS kopieren
		ld	de, CCP
		ld	hl, SHADOWRAM
		ld	bc, 1600h		;Lnge CCP+BDOS
		ldir
		; Kopieren oberen Teil BIOS nach E200
		ld	hl, memubios
		ld	bc, ubiosend-ubios	;2A9h	; bis 6710
		ldir
		; Kopieren bios-Init nach 0180
		ld	hl, membinit
		ld	de, binit
		ld	bc, binitend-binit	;43Ah	; bis 6467
		ldir

		jp	binit			; start Bios-Init

;-----------------------------------------------------------------------------
membinit:	equ	$

; Adr. biosstack 

		IF BIOSVER = 'CPMZ9OK'
biosstack:	equ	membinit+43Ah	; Stackbereich	fr Upper-BIOS E200
		ELSE
biosstack:	equ	membinit+427h	; Stackbereich	fr Upper-BIOS E200
		ENDIF




;-----------------------------------------------------------------------------
;
; startmeldung,
; Initialisieren Gertetreibertabelle,
; RAM-Test,
; starten des CCP-Systems
; Register A: untere 2 Bit Laufwerk; Bit 7 = 1 Warmstart
; Bei Bit7=0 Kaltstart mit kompletter Initialisierung, es
; kann ein Kommando beim start des CCP ausgefhrt werden
;
;-----------------------------------------------------------------------------

		phase 	0180h


binit:
		IF BIOSVER = 'CPMZ9OK'
		push	af
		and	80h
		ld	(binit7+1), a
		pop	af
		ENDIF

		and	3
		ld	(3), a		; Standard-Laufwerk
		ld	(wboot2+1), a

		ld	hl, start	; 463H Bytes ab	5FF0-6453 mit FF fllen
		ld	de, start+1	; d.h. die originale Lokation dieses Codes berschreiben
		ld	bc, biosstack-start-1

		ld	(hl), 0FFh
		ldir

		ld	hl, 0FCB0h	; Z9001-OS: Tabelle der Interruptadressen
		ld	bc, 0Ch
		ld	de, intvectab	; nach E700 kopieren
		ldir

		ld	bc, 0FFh-0Ch	; bis E7FF mit 0 fllen
		ld	h, d		; DE=E70D
		ld	l, e
		inc	de
		ld	(hl), 0
		ldir

		ld	c, 10010100b	; I/O-Byte: LIST=UL:,PUNCH=UP1:,READER=UR1:,CONST=TTY:
		call	0F03Ch		; Z9001-OS: Setzen I/O-Byte

		ld	hl, intvectab	; Interrupttabelle
		ld	a, h
		ld	i, a		; Interruptregister setzen
		im	2

; Initialisierung FDC

	IF BIOSVER <> 'CPMZ6010'

		ld	a, 20h		; 0010 0000 FDC	Reset
		out	(FDCZ),	a
		ld	a, 13h		; 0001 0011 Terminal Count aktivieren, Motor LW	1+2 an
		out	(FDCZ),	a
	ENDIF
;

binit1:		ld	b, 0		;INITIALISIERUNG P8272
binit2:		djnz	binit2
		in	a, (FDCD)
		cp	80h
		jr	z, binit4
		in	a, (FDCC)
binit3:		jr	binit1
;
stab:
		IF BIOSVER = 'CPMZ9OK'
		db  	10011111b	;XXXX=Schrittratenzeit SRT,XXXX= Kopfladezeit HUT
		db  	00111111b	;XXXXXXX=Kopfladezeit HLT,X=no dma ND
		ELSEIF BIOSVER = 'CPMSS4R'
		db  	11101111b	;XXXX=Schrittratenzeit SRT,XXXX= Kopfladezeit HUT
		db  	00111111b	;XXXXXXX=Kopfladezeit HLT,X=no dma ND
		ELSEIF BIOSVER = 'CPMZ6010'
		db 	11101111b
		db 	11111111b
		ENDIF
;
binit4:		ld	hl, stab-1	;PARAMETER LADEN

		ld	bc, 303h	;SPECIFY-COMM 3BYTES
		call	wcom1		;SCHREIBEN COMM
		xor	a
		ld	(UNIT), a
		ld	(dFDCZ1), a
		call	recal2
		call	sds
		bit	4, a
		ld	a, 1
		jr	z, binit5
		ld	(dFDCZ1), a
binit5:		ld	(UNIT), a
		call	recal2
		call	sds
		bit	4, a
		ld	a, (dFDCZ1)


		jr	z, binit6

		IF BIOSVER <> 'CPMZ6010'
		or	2
		ELSE
		inc	a
		ENDIF
		ld	(dFDCZ1), a
binit6:

		IF BIOSVER <> 'CPMZ6010'
		out	(FDCZ), a	; ZusatzRegister FDC schreiben
		ENDIF

		; Gertetreibertabellen fllen
		call	fiotab		; Gertetreibertabelle READER setzen
		dw	8
		dw	0F8F1h		; TTY:=	CRT-Treiber Z9001
;;		dw	_dummyin	; RDR:
                dw 	_UR6010		; RDR = S6010
		dw	_dummyin	; UR1:
		dw	_dummyin	; UR2:


;-----------------------------------------------------------------------------
; S6010-Treiber UBIOS Initialisierung
;-----------------------------------------------------------------------------

IS61010:        ld      hl, 0
                ld      (UR6010_b1), hl	; init.
        	jr      IS6010_1

IS6010_tab:     
		; CTC 
		db    7			; Steuerwort: DI, Zeitgeber, Zeitkonstante folgt, Reset
                db    8			; Zeitkonstante: 8 => 1200 Baud (ZKCTC=9600/BDRAT)
		
		; SIO
                db    4			; next Register WR4
                db  4Ch			; x1 Taktrate, 8 Bit-Sync-Zeichen, 2 Stop-Bit, keine Paritt
                db    1			; next Register WR1
                db    0
                db    3			; next Register WR3
                db 0E1h			; Rx 8 Bits/Zeichen, Auto Enables, Empfnger-Freigabe
                db    5			; next Register WR5
                db 0E8h			; DTR, Tx 8 Bits/Zeichen, Senderfreigabe


IS6010_1:       ld      hl, IS6010_tab
                ld      a, 18h		; SIO Reset
                out     (SIOKD), a
                ld      bc, 2*100h+CTC	; B=2 C=CTC
                otir
                ld      bc, 8*100h+SIOKD	; B=7 C=SIOKD
                otir

; ---------------------------------------------------------------------------

		call	fiotab		; Gertetreibertabelle PUNCH setzen
		dw	10h
		dw	0F8F1h		; TTY:=	CRT-Treiber Z9001
;;		dw	_pun		; PUN:
                dw 	_UP6010		; PUN = S6010
		dw	_pun		; UP1:
		dw	_up2		; UP2:

		call	fiotab		; Gertetreibertabelle LIST setzen
		dw	18h
		dw	0F8F1h		; TTY:=	CRT-Treiber Z9001
;;		dw	_CRTL		; CRT:
;;		dw	_up2		; LST:
                dw 	_up2
                dw 	_UP6010		; LST = S6010
		dw	_UL		; UL:

		xor	a
		ld	(_0_572Fh), a	; Unklar, veraltet????

		call	fiotab		; Gertetreiber	CONST setzen
		dw	0
		dw	_TTYC		; TTY:
		dw	_CRTC		; CRT:
		dw	_BATC		; BAT:
		dw	0F8F1h		; RDR:=	CRT-Treiber Z9001

		xor	a
		ld	(lstflag), a

binit7:		IF BIOSVER = 'CPMZ9OK'
		ld	a, 0
		ENDIF

		ld	hl, loc_0_E350
		ld	(off_0_E34E), hl
		ld	hl, 0
		ld	(loc_0_E350), hl
		ld	a, 0FFh
		ld	(byte_0_E34D), a
		ld	hl, ustart
		ld	(sbios+1), hl

		IF BIOSVER = 'CPMZ9OK'
		ld	hl, bootmsg	; Anzeige Boot-Meldung
		ELSEIF BIOSVER <> 'CPMZ9OK'
		ld	hl, (schutzbyte3)
		ENDIF

		call	uprnst		; Stringausgabe
		ld	hl, aRamTest	; "\r\n- RAM-Test ... "
		call	uprnst		; Anzeige RAM-Test...

		; RAM-Test
		ld	hl, BDOS	; Beginn BDOS
		ld	de, binitend+3	; Ende INIT
		or	a
		sbc	hl, de
		ld	b, h
		ld	c, l
		ex	de, hl
		di
		out	(rambank1), a	; ShadowRAM deaktivieren
		push	af
		ld	a, 'U'
		ld	(bmode), a
		pop	af
binit8:		ld	a, (hl)		; RAM-Test
		cpl
		ld	(hl), a
		cp	(hl)
		jr	nz, binit9	; Bei RAM-Fehler
		cpl
		ld	(hl), a
		inc	hl
		dec	bc
		ld	a, b
		or	c
		jr	nz, binit8
		di
		out	(rambank2), a	; ShadowRAM aktivieren
		push	af
		xor	a
		ld	(bmode), a
		pop	af

		ld	hl, aTpaIstOk	; "TPA ist OK!\r"
		call	uprnst		; Stringausgabe

		IF BIOSVER <> 'CPMZ9OK'
		call	copyprot	; Copyrightschutztest
		ENDIF
		jp	binit10

;-----------------------------------------------------------------------------
;
;-----------------------------------------------------------------------------

; HL als Hexwert nach (DE) schreiben, DE 4x inkrementieren
convhl:		ld	a, h
		call	conva		; A als	Hexwert	nach (DE) schreiben, DE	2x inkrementieren
		ld	a, l
; A als	Hexwert	nach (DE) schreiben, DE	2x inkrementieren
conva:		call	conva1
conva1:		rrca
		rrca
		rrca
		rrca
		push	af
		and	0Fh
		sub	0Ah
		jr	c, conva2
		add	a, 7
conva2:		add	a, 3Ah
		ld	(de), a
		inc	de
		pop	af
		ret

;-----------------------------------------------------------------------------
; Copyrightschutztest
;-----------------------------------------------------------------------------

		IF BIOSVER <> 'CPMZ9OK'

; Aufruf am Ende von binit
copyprot:	ld	b, 21h 			; Lnge von "implementiert von F.Schwarzenberg"
		ld	a, (schutzbyte1)
		ld	l, a
		ld	e, a
		ld	a, (schutzbyte4)
		ld	d, a			; de = ec28h = Adr. zweite Bildschirmzeile
		sub	4
		ld	h, a			; hl = e828h = Adr. zweite Farbzeile
		ld	a, (de)			; Prfwert: startwert 1. Buchstabe
copyprot1:	rrca				; Prfwert rotieren
		ld	(hl), 2			; Hintergrundfarbe ndern auf ...
		inc	hl
		inc	de
		ex	de, hl
		xor	(hl)			; Prfwert mit nchstem Buchstaben verknpfen
		ex	de, hl
		djnz	copyprot1
		ld	(schutzbyte2), a	; mu 0ACh ergeben, s. a. binit10
		ret

		ENDIF

;-----------------------------------------------------------------------------
;
;-----------------------------------------------------------------------------
; Wenn Fehler beim Ramtest
binit9:		di
		out	(rambank2), a	; ShadowRAM aktivieren
		push	af
		xor	a
		ld	(bmode), a
		pop	af
		ld	de, aRamFehler+1Bh
		call	convhl		; HL als Hexwert nach (DE) schreiben, DE 4x inkrementieren
		ld	de, BDOS	; Adresse BDOS
		dec	hl
		ld	(hl), d
		dec	hl
		ld	(hl), e
		dec	hl
		ld	(hl), 0C3h	; 'JP'
		ld	(wboot0+1), hl	; Adresse BDOS; wird zu CALL 5 kopiert
		dec	hl
		ld	de, aRamFehler+30h
		call	convhl		; HL als Hexwert nach (DE) schreiben, DE 4x inkrementieren

		ld	hl, aRamFehler	; "\r\n!!RAM-Fehler auf Adresse XXXXH!!  TPA"...
		call	uprnst		; Stringausgabe

		IF BIOSVER = 'CPMZ9OK'

binit10:	ld	hl, CCP+0003h	; CLEAR: CCP-Aufruf mit	Lschen	des Befehlsbuffers
		ld	a, (binit7+1)	; Register AF vom Aufruf
		or	a
		jr	nz, binit11

		ELSEIF	BIOSVER <> 'CPMZ9OK'

binit10:	ld	a, (schutzbyte2)
		cp	0ACh
		jr	z, binit12	; alles OK, Impl. Zeile wurde nicht verndert

		; Copyright-nderung ahnden: Speicher lschen !
		ld	hl, 0B0EDh	; Auf Adr. 80 Code bringen:
		ld	(80h), hl	; 0080: ED B0	ldir
		ld	hl, 0FC18h	; 0082: 18 FC	jr zu Adr. 0080H
		ld	(82h), hl	; 0084: 23	Zeichen '#'
		ld	hl, 0084h
		ld	d, h
		ld	e, l
		inc	de
		ld	(hl), 23h

		ld	bc, 0E800h-0084h	; hl=0084h, de = 0085h, bc=0E77Ch
		jp	80h		; Speicher von 84h-E7FFh mit '#' fllen

		ENDIF

binit12:	ld	hl, startkdo	; start-Kommando
		ld	de, CCP+0007h	; CCP-Befehlsbuffer
		ld	bc, 81h		; 80 Byte + 1 Lngenbyte
		ldir
		ld	hl, CCP		; COLD:	CCP-start

		IF BIOSVER = 'CPMZ9OK'
binit11:	jp	wboot		; WBOOT	im ShadowBios
		ELSEIF BIOSVER <> 'CPMZ9OK'
binit11:	jp	wboot3		; WBOOT	im ShadowBios, mit Ausfhrung des startkdo :-))
		ENDIF


;-----------------------------------------------------------------------------
; Stringausgabe
;-----------------------------------------------------------------------------

		IF BIOSVER = 'CPMZ9OK'

uprnst:		ld	a, (binit7+1)	; Register AF vom Aufruf
		or	a
		ret	nz
		jp	prnst		; shadow prnst

		ELSEIF	BIOSVER <> 'CPMZ9OK'

uprnst:		equ	prnst

		ENDIF

;-----------------------------------------------------------------------------
; Gertetreibertabelle fllen
; Nach Call 1 Word Offset, dann	4 Word Adressen	der Gertetreiber
;-----------------------------------------------------------------------------
fiotab:		pop	hl		; Adresse nach Call holen (Ret-Adr. vom	Stack)
		ld	e, (hl)
		inc	hl
		ld	d, (hl)
		inc	hl
		ex	de, hl		; HL=Offset
		ld	bc, 0EFC9h	; Tabelle der Gertetreiberadressen
		add	hl, bc		; Offset addieren
		ex	de, hl
		ld	bc, 8
		ldir			; 4 Word kopieren
		jp	(hl)		; und zurck zum Aufrufer

bootmsg:	db  	0Ch
		IF BIOSVER = 'CPMZ9OK'

		db	 " CP/A Z9001, Version 18.01.89", 0Ah, 0Dh
		db  	16h, "implementiert von F.Schwarzenberg", 16h, 0Ah, 0Dh
		db	"TPA 100H - 0D405H", 0Dh, 0Ah
		db	"- Disks: 5", '"', "(80,DS)/5", '"', "(80,DS)", 0Dh, 0Ah

		ELSEIF	BIOSVER = 'CPMSS4R'

		db	 " CP/A Z9001, Version 18.03.89", 0Ah, 0Dh
		db  	16h, "implementiert von F.Schwarzenberg", 16h, 0Ah, 0Dh ; der Text wird im Copyrightschutztest geprft!!
		db	"TPA 100H - 0D405H", 0Dh, 0Ah
		db	"- Disks: 5", '"', "(40,SS)/5", '"', "(40,SS)", 0Dh, 0Ah

		ELSEIF	BIOSVER = 'CPMZ6010'

		db	 " CP/A Z9001, Version 23.02.89", 0Ah, 0Dh
		db  	16h, "implementiert von F.Schwarzenberg", 16h, 0Ah, 0Dh ; der Text wird im Copyrightschutztest geprft!!
		db	"TPA 100H - 0D405H", 0Dh, 0Ah
		db	"- Disks: 5", '"', "(80,DS)/5", '"', "(80,DS)", 0Dh, 0Ah

		ENDIF
		db 	"IO-Byte:", 0Dh, 0Ah

;;		db	"Bit 7,6 (LST:-Kanal):", 0Dh, 0Ah
;;		db	"     00 (TTY:): ", 0Dh, 0Ah
;;		db	"     01 (CRT:): V24-User-Port 1200 Bd", 0Dh, 0Ah
;;		db	"     10 (LPT:): V24-User-Port 9600 Bd", 0Dh, 0Ah
;;		db	"     11 (UL1:): Centronics User-Port", 0Dh, 0Ah
;;		db	"Bit 5,4 (PUN:-Kanal):", 0Dh, 0Ah
;;		db	"Bit 3,2 (RDR:-Kanal):", 0Dh, 0Ah
;;		db	"Bit 1,0 (CON:-Kanal):", 0Dh, 0Ah
;;		db	"     00 (TTY:): Beep + Gross<=>Klein", 0Dh, 0Ah
;;		db	"     01 (CRT:): nur  Tastatur-Beep", 0Dh, 0Ah
;;		db	"     11 (UC1:): ohne Beep"
;;		db 	0

                db 	"Bit 7,6 (LST:-Kanal):",0Dh,0Ah
                db 	"     00 (TTY:): ",0Dh,0Ah
                db 	"     01 (CRT:): V24-User-Port 9600 Bd",0Dh,0Ah
                db 	"     10 (LPT:): S 6010 (SIO)  1200 Bd",0Dh,0Ah
                db 	"     11 (UL1:): Centronics User-Port",0Dh,0Ah
                db 	"Bit 5,4 (PUN:-Kanal):",0Dh,0Ah
                db 	"     01 (PTP:): S 6010 (SIO)  1200 Bd",0Dh,0Ah
                db 	"Bit 3,2 (RDR:-Kanal):",0Dh,0Ah
                db 	"     01 (PTR:): S 6010 (SIO)  1200 Bd",0Dh,0Ah
                db 	"Bit 1,0 (CON:-Kanal):",0Dh,0Ah
                db 	"     00 (TTY:): Beep + Gross<=>Klein",0Dh,0Ah
                db 	"     01 (CRT:): nur  Tastatur-Beep",0Dh,0Ah
                db 	"     11 (UC1:): ohne Beep",0

aRamTest:	db 	0Dh, 0Ah, "- RAM-Test ... " , 0

aTpaIstOk:	db	"TPA ist OK!", 0Dh, 0

aRamFehler:	db 	0Dh, 0Ah, "!!RAM-Fehler auf Adresse XXXXH!!  TPA nur bis YYYYH.", 7, 0Dh, 0Ah, 0

; hier kann ein	Befehl eingetragen werden, der beim start des CCP ausgefhrt werden soll
startkdo:	db  	7Fh	; Lnge	des Befehls(buffers)
;;		IF BIOSVER = 'CPMSS4R'
		db 	"CLOCK"
;;		ENDIF
		db	128-($-startkdo-1) dup (0)	; mit 00 auffllen

binitend:	equ	$

		dephase


;-----------------------------------------------------------------------------
memubios:	equ	$












;-----------------------------------------------------------------------------
;
; Upper BIOS	E200h - E519h
;
;-----------------------------------------------------------------------------

		phase 	UBIOS

; Sprungvektortabelle
		jp	_boot
		jp	_wboot
		jp	_const
		jp	0F009h		; Z9001-OS: conin
		jp	_conout
		jp	0F00Fh		; Z9001-OS: list
		jp	0F012h		; Z9001-OS: punch
		jp	0F015h		; Z9001-OS: reader
		jp	_home
		jp	_seldsk
		jp	_settrk
		jp	_setsec
		jp	_setdma
		jp	_read
		jp	_write
		jp	0F02Dh		; Z9001-OS: listst
		jp	_sectran

;korrespondierende Umsetzungstabelle fr Aufuf der Schattenbios-Funktionen
_boot:		call	csbios		; Aufruf im Schattenbios
_wboot:		call	csbios		; Aufruf im Schattenbios
		call	csbios		; Aufruf im Schattenbios
		call	csbios		; Aufruf im Schattenbios
		call	csbios		; Aufruf im Schattenbios
		call	csbios		; Aufruf im Schattenbios
		call	csbios		; Aufruf im Schattenbios
		call	csbios		; Aufruf im Schattenbios
_home:		call	csbios		; Aufruf im Schattenbios
_seldsk:	call	csbios		; Aufruf im Schattenbios
_settrk:	call	csbios		; Aufruf im Schattenbios
_setsec:	call	csbios		; Aufruf im Schattenbios
_setdma:	call	csbios		; Aufruf im Schattenbios
_read:		jp	_read1
_write:		jp	_write1
		call	csbios		; Aufruf im Schattenbios
_sectran:	call	csbios


;-----------------------------------------------------------------------------
; call shadow bios function
;-----------------------------------------------------------------------------

csbios:		pop	hl		; HL = Returnadr. vom Stack = E236 (_wboot)  ...
		push	de		; DE sichern
		ld	de, _wboot
		or	a
		sbc	hl, de
		ld	de, sbios	; Adr. Schattenbios
		add	hl, de		; HL = orig.Return - adr(_wboot) + adr(sbios)
					; damit	HL = Adr. im Schattenbios
		pop	de		; DE restaurieren
csbios1:	ld	(stackp), sp	; Zwischenspeicher Stackpointer
		di
		out	(rambank2), a	; Schattenbios aktiv
		push	af
		xor	a
		ld	(bmode), a	; Bios-Mode: 'U'-Arbeit	im Upper Bios, 00-Arbeit im Schatten-Bios
		pop	af
		ei
		ld	sp, biosstack	; Adr. BIOS-Stackbereich
		ld	(csbios2+1), hl	; Memory Patch nachfolgender CALL-Befehl
csbios2:	call	0		; Call berechnete Adresse (HL)
		ld	sp, (stackp)	; Zwischenspeicher Stackpointer
		di
		out	(rambank1), a	; Schattenbios wegschalten
		push	af
		ld	a, 'U'
		ld	(bmode), a	; Bios-Mode: 'U'-Arbeit	im Upper Bios, 00-Arbeit im Schatten-Bios
		pop	af
		ei
		or	a
		ret

;-----------------------------------------------------------------------------
; wboot beenden und Bergabe der Kontrolle ans CCP
;-----------------------------------------------------------------------------

wboot3u:	di
		out	(rambank1), a	; ShadowRAM deaktivieren
		push	af
		ld	a, 'U'
		ld	(bmode), a	; Bios-Mode: 'U'-Arbeit	im Upper Bios, 00-Arbeit im Schatten-Bios
		pop	af
		ei
wboot3u1:	jp	CCP+0003h	; CLEAR: CCP-Aufruf mit	Lschen	des Befehlsbuffers

;-----------------------------------------------------------------------------
; write disk
;-----------------------------------------------------------------------------
_write1:	ld	hl, (ddma)
		ld	de, rwbuffer
		push	bc
		ld	bc, 80h
		ldir
		pop	bc
		ld	hl, uwrite	; WRITE	im Schattenbios
		jr	csbios1

;-----------------------------------------------------------------------------
; read disk
;-----------------------------------------------------------------------------
_read1:		di
		out	(rambank2), a	; ShadowRAM aktivieren
		push	af
		xor	a
		ld	(bmode), a	; Bios-Mode: 'U'-Arbeit	im Upper Bios, 00-Arbeit im Schatten-Bios
		pop	af
		ei
		ld	(stackp), sp	; Zwischenspeicher Stackpointer
		ld	sp, biosstack	; Adr. BIOS-Stackbereich
		call	uread		; READ im Schattenbios
		ld	sp, (stackp)	; Zwischenspeicher Stackpointer
		di
		out	(rambank1), a	; ShadowRAM deaktivieren
		push	af
		ld	a, 'U'
		ld	(bmode), a	; Bios-Mode: 'U'-Arbeit	im Upper Bios, 00-Arbeit im Schatten-Bios
		pop	af
		ei
		ld	hl, rwbuffer
		ld	de, (ddma)
		ld	bc, 80h
		ldir
		or	a
		ret

;-----------------------------------------------------------------------------
; console status
;-----------------------------------------------------------------------------
_const:		call	0F006h		; Z9001-OS: Abfrage Status CONST
		or	a
		ret	z
		ld	a, 0FFh
		ret

;-----------------------------------------------------------------------------
; BAT-Gertereiber fr CRT
;-----------------------------------------------------------------------------
_BATC:		or	a
		jp	nz, 0F7B4h	; Z9001-OS: BAT, Steuerprogramm fr Batch-Mode von CONST
		ld	hl, (0F016h)	; Z9001-OS: Adresse des UP READER (Eingabe Zeichen von Reader)
		inc	hl
		inc	hl
		jp	(hl)

;-----------------------------------------------------------------------------
; Dummy-Gertetreiber, gibt stets Dateiende (^Z) zurck
;-----------------------------------------------------------------------------
_dummyin:				; ^Z; Eingabeende
		ld	a, 1Ah
		ret

;-----------------------------------------------------------------------------
; console character out
;-----------------------------------------------------------------------------
_conout:	ld	a, (byte_0_E34D)
		cp	0FFh
		jr	z, _conout3
		ld	hl, (off_0_E34E)
		res	7, c
		inc	c
		ld	(hl), c
		dec	hl
		ld	(off_0_E34E), hl
		ld	hl, byte_0_E34D
		dec	(hl)
		jr	nz, _conout2
		dec	(hl)
		ld	hl, _conout1
		push	hl
		ld	hl, (0F069h)	; Z9001-OS: Adr. Setzen log. Cursor
		push	hl
		ld	l, 3
		ld	de, (loc_0_E350)
		ret
_conout1:	ld	hl, loc_0_E351
		ld	(off_0_E34E), hl
_conout2:	or	a
		ret
_conout3:	ld	a, (25h)	; KEYBU	(Tastaturbuffer)
		cp	13h		; PAUSE-Taste?
		jr	nz, _conout4
		push	bc
		call	0F009h		; Z9001-OS: Eingabe Zeichen von CONST, PAUSE-Taste holen
		call	0F009h		; Z9001-OS: Eingabe Zeichen von CONST, Warten auf Tastendruck
		pop	bc
_conout4:	ld	a, 1Bh
		cp	c
		jp	nz, 0F00Ch	; Z9001-OS: Ausgabe Zeichen zu CONST
		ld	a, 2
		ld	(byte_0_E34D), a
		jr	_conout1

byte_0_E34D:	db 	0FFh
off_0_E34E:	dw	loc_0_E350
loc_0_E350:	db	0
loc_0_E351:	db	0

bmode:		db 	0		; Bios-Mode: 'U'-Arbeit	im Upper Bios, 00-Arbeit im Schatten-Bios

;-----------------------------------------------------------------------------
; TTY-Treiber fr CONS
;-----------------------------------------------------------------------------
_TTYC:		ld	l, a
		ld	a, (bmode)	; Bios-Mode: 'U'-Arbeit	im Upper Bios, 00-Arbeit im Schatten-Bios
		or	a
		ld	a, l
		jp	z, uttyc
		di
		out	(rambank2), a	; ShadowRAM aktivieren
		push	af
		xor	a
		ld	(bmode), a	; Bios-Mode: 'U'-Arbeit	im Upper Bios, 00-Arbeit im Schatten-Bios
		pop	af
		ei
		call	uttyc
		di
		out	(rambank1), a	; ShadowRAM deaktivieren
		push	af
		ld	a, 'U'
		ld	(bmode), a	; Bios-Mode: 'U'-Arbeit	im Upper Bios, 00-Arbeit im Schatten-Bios
		pop	af
		ei
		ret

;-----------------------------------------------------------------------------
; CRT-Treiber fr CONS
;-----------------------------------------------------------------------------
_CRTC:		ld	l, a
		ld	a, (bmode)	; Bios-Mode: 'U'-Arbeit	im Upper Bios, 00-Arbeit im Schatten-Bios
		or	a
		ld	a, l
		jp	z, ucrtc
		di
		out	(rambank2), a	; ShadowRAM aktivieren
		push	af
		xor	a
		ld	(bmode), a	; Bios-Mode: 'U'-Arbeit	im Upper Bios, 00-Arbeit im Schatten-Bios
		pop	af
		ei
		call	ucrtc
		di
		out	(rambank1), a	; ShadowRAM deaktivieren
		push	af
		ld	a, 'U'
		ld	(bmode), a	; Bios-Mode: 'U'-Arbeit	im Upper Bios, 00-Arbeit im Schatten-Bios
		pop	af
		ei
		ret

;-----------------------------------------------------------------------------
; UP2 Gertreiber fr PUNCH
;-----------------------------------------------------------------------------
_up2:		ld	l, a
		ld	a, (bmode)	; Bios-Mode: 'U'-Arbeit	im Upper Bios, 00-Arbeit im Schatten-Bios
		or	a
		ld	a, l
		jp	z, uup
		di
		out	(rambank2), a	; ShadowRAM aktivieren
		push	af
		xor	a
		ld	(bmode), a	; Bios-Mode: 'U'-Arbeit	im Upper Bios, 00-Arbeit im Schatten-Bios
		pop	af
		ei
		call	uup
		di
		out	(rambank1), a	; ShadowRAM deaktivieren
		push	af
		ld	a, 'U'
		ld	(bmode), a	; Bios-Mode: 'U'-Arbeit	im Upper Bios, 00-Arbeit im Schatten-Bios
		pop	af
		ei
		ret

;-----------------------------------------------------------------------------
; CRT-Gertetreiber fr	LIST
;-----------------------------------------------------------------------------
_CRTL:		ld	l, a
		ld	a, (bmode)	; Bios-Mode: 'U'-Arbeit	im Upper Bios, 00-Arbeit im Schatten-Bios
		or	a
		ld	a, l
		jp	z, ucrtl
		di
		out	(rambank2), a	; ShadowRAM aktivieren
		push	af
		xor	a
		ld	(bmode), a	; Bios-Mode: 'U'-Arbeit	im Upper Bios, 00-Arbeit im Schatten-Bios
		pop	af
		ei
		call	ucrtl
		di
		out	(rambank1), a	; ShadowRAM deaktivieren
		push	af
		ld	a, 'U'
		ld	(bmode), a	; Bios-Mode: 'U'-Arbeit	im Upper Bios, 00-Arbeit im Schatten-Bios
		pop	af
		ei
		ret

;-----------------------------------------------------------------------------
; UBIOS S6010-Treiber f. PUNCH
;-----------------------------------------------------------------------------


_UP6010:        ld      l, a
                ld      a, (bmode)	; Bios-Mode: 'U'-Arbeit	im Upper Bios, 00-Arbeit im Schatten-Bios
                or      a
                ld      a, l
                jp      z, UP6010
                di
                out     (rambank2), a	; ShadowRAM aktivieren
                push    af
                xor     a
                ld      (bmode), a	; Bios-Mode: 'U'-Arbeit	im Upper Bios, 00-Arbeit im Schatten-Bios
                pop     af
                ei
                call    UP6010
                di
                out     (rambank1), a	; ShadowRAM deaktivieren
                push    af
                ld      a, 'U'
                ld      (bmode), a	; Bios-Mode: 'U'-Arbeit	im Upper Bios, 00-Arbeit im Schatten-Bios
                pop     af
                ei
                ret

;-----------------------------------------------------------------------------
; UBIOS S6010-Treiber f. READER
;-----------------------------------------------------------------------------

_UR6010:        ld      l, a
                ld      a, (bmode)	; Bios-Mode: 'U'-Arbeit	im Upper Bios, 00-Arbeit im Schatten-Bios
                or      a
                ld      a, l
                jp      z, UR6010
                di
                out     (rambank2), a	; ShadowRAM aktivieren
                push    af
                xor     a
                ld      (bmode), a	; Bios-Mode: 'U'-Arbeit	im Upper Bios, 00-Arbeit im Schatten-Bios
                pop     af
                ei
                call    UR6010
                di
                out     (rambank1), a	; ShadowRAM deaktivieren
                push    af
                ld      a, 'U'
                ld      (bmode), a	; Bios-Mode: 'U'-Arbeit	im Upper Bios, 00-Arbeit im Schatten-Bios
                pop     af
                ei
                ret

;-----------------------------------------------------------------------------
; UL Gertetreiber fr LIST
;-----------------------------------------------------------------------------
_UL:		ld	l, a
		ld	a, (bmode)	; Bios-Mode: 'U'-Arbeit	im Upper Bios, 00-Arbeit im Schatten-Bios
		or	a
		ld	a, l
		jp	z, uul
		di
		out	(rambank2), a	; ShadowRAM aktivieren
		push	af
		xor	a
		ld	(bmode), a	; Bios-Mode: 'U'-Arbeit	im Upper Bios, 00-Arbeit im Schatten-Bios
		pop	af
		ei
		call	uul
		di
		out	(rambank1), a	; ShadowRAM deaktivieren
		push	af
		ld	a, 'U'
		ld	(bmode), a	; Bios-Mode: 'U'-Arbeit	im Upper Bios, 00-Arbeit im Schatten-Bios
		pop	af
		ei
		ret

;-----------------------------------------------------------------------------
; PUN-Gertetreiber fr	PUNCH
;-----------------------------------------------------------------------------
_pun:		ld	l, a
		ld	a, (bmode)	; Bios-Mode: 'U'-Arbeit	im Upper Bios, 00-Arbeit im Schatten-Bios
		or	a
		ld	a, l
		jp	z, upunch
		di
		out	(rambank2), a	; ShadowRAM aktivieren
		push	af
		xor	a
		ld	(bmode), a	; Bios-Mode: 'U'-Arbeit	im Upper Bios, 00-Arbeit im Schatten-Bios
		pop	af
		ei
		call	upunch
		di
		out	(rambank1), a	; ShadowRAM deaktivieren
		push	af
		ld	a, 'U'
		ld	(bmode), a	; Bios-Mode: 'U'-Arbeit	im Upper Bios, 00-Arbeit im Schatten-Bios
		pop	af
		ei
		ret


;-----------------------------------------------------------------------------
; CPA-Disk-Parameter
;-----------------------------------------------------------------------------
;Disk-Parameter-Header
dph0:		dw	0		; XLT	translation table
		dw	0
		dw	0
		dw	0
		dw	dirbuf		; DIRBUF
		dw	dpb0		; DPB
		dw	CSV0		; CSV
		dw	ALV0		; ALV

dph1:		dw	0		; XLT	translation table
		dw	0
		dw	0
		dw	0
		dw	dirbuf		; DIRBUF
		dw	dpb1  		; DPB
		dw	CSV1		; CSV
		dw	ALV1		; ALV

dphtab:		dw	dph0
		dw	dph1

		IF BIOSVER <> 'CPMSS4R'

;Disp-Parameter-Blcke
;800K-Laufwerk 80 Spuren DS, DD
dpb0:		dw	40		; SPT	40 Sektoren/Spur
		db 	4		; BSH	= log2(blocksize/128)
					;  => blocksize = (2^4)*128 = 2048
		db 	00001111b	; BLM	= 2^BSH-1
		db 	0		; EXM
		dw	400-1		; DSM	max. 400 Blcke
					;  => also 400*2048 = 800K disc
		dw	192-1		; DRM	max 192 Dir. Eintrge
		db 	11100000b	; AL0	also 3 Blcke (192*32 / 2048)
		db 	0		; AL1	fr DIR reservieren
		dw	48		; CKS	= (DRM+1)/4
		dw	0		; OFF	0 reservierte Spuren
;Erweiternung DBP des Z9001-CBIOS
		db  	00100000b 	; DPB-Steuerinformationen
					; bit 5=1 Double sided zu betreiben
		db    	3 		; Sektorlaengencode
		db    	7 		; Puffermaske 8*128
		db    	1 		; Zahl der Stepimpulse
		db    	0  		; keine Weiter-Numm. auf Rueckseite
		db  	80		; benutzte Zahl phys. Spuren
		db    	0  		; (dbp+15h) phys. LW-Nr. (0 --> A:)
		db  	80		; Gesamtzahl physische Spuren (wird in FORMATZ abgefragt)
		db	4*10		; Schrittzeit fuer 80 Tr. DS LW
		db	00111001b	; vom Laufwerkstyp abhaengige Flags
					; bit 0=0 bei SS, =1 bei DS
					; bit 3=0 bei SD (FM), =1 bei DD (MFM)
					; bit 4=0 bei 8", =1 bei 5"
					; bit 5=0 bei nicht 80-Spur, =1 bei 80-Spur
					; bit 6=0 ohne, =1 mit Verify nach Schreiben
					; bit 7=0 bei max 26 ,=1 bei max 52 Sekt./Spur (8" MFM)
;translation tab dbp0
		db	1,2,3,4,5,6,7,8,9,10,11,12,13,14
		db	15,16,17,18,19,20,21,22,23,24,25,26

dpb1:		dw	40		; SPT	40 Sektoren/Spur
		db 	4               ; BSH	= log2(blocksize/128)
                                        ;  => blocksize = (2^4)*128 = 2048
		db 	00001111b       ; BLM	= 2^BSH-1
		db 	0               ; EXM
		dw	400-1           ; DSM	max. 400 Blcke
                                        ;  => also 400*2048 = 800K disc
		dw	192-1           ; DRM	max 192 Dir. Eintrge
		db 	11100000b       ; AL0	also 3 Blcke (192*32 / 2048)
		db 	0               ; AL1	fr DIR reservieren
		dw	48              ; CKS	= (DRM+1)/4
		dw	0               ; OFF	0 reservierte Spuren
; CPA-Erweiterung
		db  	00100000b 	; DPB-Steuerinformationen
					; Bit5=1 Double sided zu betreiben
		db    	3 		; Sektorlaengencode
		db    	7 		; Puffermaske 8*128
		db    	1 		; Zahl der Stepimpulse
		db    	0  		; keine Weiter-Numm. auf Rueckseite
		db  	80		; benutzte Zahl phys. Spuren
		db    	1  		; (dbp+15h) phys. LW-Nr. (1 --> B:)
		db  	80		; Gesamtzahl physische Spuren (wird in FORMATZ abgefragt)
		db	4*10		; Schrittzeit fuer 80 Tr. DS LW
		db	00111001b	; vom Laufwerkstyp abhaengige Flags
					; bit 0=0 bei SS, =1 bei DS
					; bit 3=0 bei SD (FM), =1 bei DD (MFM)
					; bit 4=0 bei 8", =1 bei 5"
					; bit 5=0 bei nicht 80-Spur, =1 bei 80-Spur
					; bit 6=0 ohne, =1 mit Verify nach Schreiben
					; bit 7=0 bei max 26 ,=1 bei max 52 Sekt./Spur (8" MFM)
; translation tab dbp1
		db	1,2,3,4,5,6,7,8,9,10,11,12,13,14
		db	15,16,17,18,19,20,21,22,23,24,25,26

		ELSEIF	BIOSVER <> 'CPMZ9OK'

;200K-Laufwerk: 40 Spuren SS, SD
dpb0:		dw   40			; SPT	40 Sektoren/Spur
		db    3                 ; BSH	= log2(blocksize/128)
					;  => blocksize = (2^3)*128 = 1024
		db    00000111b         ; BLM	= 2^BSH-1
		db    0                 ; EXM
		dw 200-1                ; DSM	max. 200 Blcke
					;  => also 200*1024 = 200K disc
		dw  64-1                ; DRM	max 64 Dir. Eintrge
		db 11000000b            ; AL0	also 2 Blcke (64*32 / 1024)
		db    0                 ; AL1	fr DIR reservieren
		dw   16                 ; CKS	= (DRM+1)/4
		dw    0                 ; OFF	0 reservierte Spuren
;
		db      00000000b	; DPB-Steuerinformationen
					; bit 5=1 Double sided zu betreiben
		db    	3 		; Sektorlaengencode
		db    	7 		; Puffermaske 8*128
		db    	1 		; Zahl der Stepimpulse
		db    	0  		; keine Weiter-Numm. auf Rueckseite
		db  	40		; benutzte Zahl phys. Spuren
		db    	0		; phys. LW-Nr. (0 --> A:)
		db  	40		; Gesamtzahl physische Spuren (wird in FORMATZ abgefragt)
		db  	13*10		; Schrittzeit fuer 80 Tr. SS LW
		db  	00011000b	; vom Laufwerkstyp abhaengige Flags
					; bit 0=0 bei SS, =1 bei DS
					; bit 3=0 bei SD (FM), =1 bei DD (MFM)
					; bit 4=0 bei 8", =1 bei 5"
					; bit 5=0 bei nicht 80-Spur, =1 bei 80-Spur
					; bit 6=0 ohne, =1 mit Verify nach Schreiben
					; bit 7=0 bei max 26 ,=1 bei max 52 Sekt./Spur (8" MFM)
;translation tab dbp0
		db	1,2,3,4,5,6,7,8,9,10,11,12,13,14
		db	15,16,17,18,19,20,21,22,23,24,25,26
;
dpb1:		dw  40			; SPT	40 Sektoren/Spur
		db    3                 ; BSH	= log2(blocksize/128)
					;  => blocksize = (2^4)*128 = 1024
		db    00000111b         ; BLM	= 2^BSH-1
		db    0                 ; EXM
		dw 200-1                ; DSM	max. 200 Blcke
					;  => also 200*1024 = 200K disc
		dw  64-1                ; DRM	max 64 Dir. Eintrge
		db 11000000b            ; AL0	also 2 Blcke (64*32 / 1024)
		db    0                 ; AL1	fr DIR reservieren
		dw   16                 ; CKS	= (DRM+1)/4
		dw    0                 ; OFF	0 reservierte Spuren
;
		db      00000000b	; DPB-Steuerinformationen
					; bit 5=1 Double sided zu betreiben
		db    	3 		; Sektorlaengencode
		db    	7 		; Puffermaske 8*128
		db    	1 		; Zahl der Stepimpulse
		db    	0  		; keine Weiter-Numm. auf Rueckseite
		db  	40		; benutzte Zahl phys. Spuren
		db    	1		; phys. LW-Nr. (1 --> B:)
		db  	40		; Gesamtzahl physische Spuren (wird in FORMATZ abgefragt)
		db  	13*10		; Schrittzeit fuer 80 Tr. SS LW
		db  	00011000b	; vom Laufwerkstyp abhaengige Flags
					; bit 0=0 bei SS, =1 bei DS
					; bit 3=0 bei SD (FM), =1 bei DD (MFM)
					; bit 4=0 bei 8", =1 bei 5"
					; bit 5=0 bei nicht 80-Spur, =1 bei 80-Spur
					; bit 6=0 ohne, =1 mit Verify nach Schreiben
					; bit 7=0 bei max 26 ,=1 bei max 52 Sekt./Spur (8" MFM)
;translation tab dbp1
		db	1,2,3,4,5,6,7,8,9,10,11,12,13,14
		db	15,16,17,18,19,20,21,22,23,24,25,26

		ENDIF

ubiosend:	equ	$
romend:		equ	$-ubios+memubios

;-----------------------------------------------------------------------------
; BIOS-RAM
;-----------------------------------------------------------------------------

stackp:		dw	?		; Zwischenspeicher Stackpointer
ddma:		dw	?		; aktuelle DMA-Adresse
dirbuf:		db	128 dup (?)	; Bereich 128 Byte

		IF BIOSVER <> 'CPMSS4R'
ALV0:		db	50 dup (?)	;0E52Dh			= DSM/8+1 = 399/8+1 = 50
CSV0:		db	64 dup (?)	;0E55Fh			= ? eigentlich CKS (also 48)
ALV1:		db	50 dup (?)	;0E59Fh
CSV1:		db	64 dup (?)	;0E5D1h
					;rwbuffer 0E611h
		ELSEIF BIOSVER = 'CPMSS4R'
ALV0:		db	25 dup (?)	;0E52Dh			= DSM/8+1 = 199/8+1 = 25
CSV0:		db	16 dup (?)	;0E546h			= CKS
ALV1:		db	25 dup (?)	;0E556h
CSV1:		db	16 dup (?)	;0E56Fh
					;rwbuffer 0E57Fh
		ENDIF

rwbuffer:	db	80h dup (?)	;128 Byte Buffer fr _read/_write

lastram:	equ	$

intvectab:	equ	0E700h		; Interruptvektortabelle, 12 Byte
; bis E7FF wird mit FFH aufgefllt

		dephase
;-----------------------------------------------------------------------------

biosend:	equ	$

		end
